'''
Manages the file format and class for correlations.

:copyright:
    The SeisMIC development team (makus@gfz-potsdam.de).
:license:
    EUROPEAN UNION PUBLIC LICENCE v. 1.2
   (https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12)
:author:
   Peter Makus (makus@gfz-potsdam.de)

Created: Friday, 16th April 2021 03:21:30 pm
Last Modified: Tuesday, 4th April 2023 05:04:21 pm
'''
import ast
import fnmatch
import os
import re
from typing import List
import warnings
from copy import deepcopy

import numpy as np
# from numpy.core.fromnumeric import compress
from obspy.core.utcdatetime import UTCDateTime
from obspy.core import Stats
import h5py

from seismic.correlate.stream import CorrStream, CorrTrace

hierarchy = "/{tag}/{network}/{station}/{channel}/{corr_st}/{corr_et}"
h5_FMTSTR = os.path.join("{dir}", "{network}.{station}.h5")


class DBHandler(h5py.File):
    """
    The actual file handler of the hdf5 correlation files.

    .. warning::

        **Should not be accessed directly. Access
        :class:`~seismic.db.corr_hdf5.CorrelationDataBase` instead.**

    Child object of :class:`h5py.File` and inherets all its attributes and
    functions in addition to functions that are particularly useful for noise
    correlations.
    """
    def __init__(self, path, mode, compression, co):
        super(DBHandler, self).__init__(path, mode=mode)
        if isinstance(compression, str):
            self.compression = re.findall(r'(\w+?)(\d+)', compression)[0][0]
            if self.compression != 'gzip':
                raise ValueError(
                    'Compression of type %s is not supported.'
                    % self.compression)
            self.compression_opts = int(
                re.findall(r'(\w+?)(\d+)', compression)[0][1])
            if self.compression_opts not in np.arange(1, 10, 1, dtype=int):
                ii = np.argmin(abs(
                    np.arange(1, 10, 1, dtype=int) - self.compression_opts))
                self.compression_opts = np.arange(1, 10, 1, dtype=int)[ii]
                warnings.warn(
                    'Chosen compression level is not available for %s. \
%s Has been chosen instead (closest)' % (
                        self.compression, str(self.compression_opts)))
        else:
            self.compression = None
            self.compression_opts = None

        # check out the processing that was done on the correlations. Different
        # data will not be allowed
        if co is not None:
            try:
                co_old = self.get_corr_options()
                if co_old != co_to_hdf5(co):
                    try:
                        diff = {k: (v, co_old[k]) for k, v in co_to_hdf5(
                            co).items() if v != co_old[k]}
                    except KeyError as e:
                        raise PermissionError(
                            f'One option is not defined in new dict. {e}'
                        )
                    raise PermissionError(
                        f'The output file {path} already exists and contains'
                        ' data with'
                        ' different processing parameters. Differences are:'
                        '\nFirst: New parameters; Second: Old parameters'
                        f'\n{diff}'
                    )
            except KeyError:
                self.add_corr_options(co)

    def _close(self):
        self.close()

    def add_corr_options(self, co: dict):
        sco = str(co_to_hdf5(co))
        ds = self.create_dataset('co', data=np.empty(1))
        ds.attrs['co'] = sco

    def add_correlation(
            self, data: CorrTrace or CorrStream, tag='subdivision'):
        """
        Add correlation data to the hdf5 file. Can be accessed using the
        :func:`~seismic.db.corr_hdf5.DBHandler.get_data()` method.

        :param data: Data to save. Either a
            :class:`~seismic.correlate.correlate.CorrTrace` object or a
            :class:`~seismic.correlate.correlate.CorrStream` holding one or
            several traces.
        :type data: CorrTrace or CorrStream
        :param tag: The tag that the data should be saved under. By convention,
            unstacked correlations are saved with the tag `'subdivision'`,
            whereas stacks are saved with the tag `stack_$stacklen$`, where
            $stacklen$ is to be replaced by the length of the stack in seconds.
        :raises TypeError: for wrong data type.
        """
        if not isinstance(data, CorrTrace) and\
                not isinstance(data, CorrStream):
            raise TypeError('Data has to be either a \
:class:`~seismic.correlate.correlate.CorrTrace` object or a \
:class:`~seismic.correlate.correlate.CorrStream` object')

        if isinstance(data, CorrTrace):
            data = [data]

        for tr in data:
            st = tr.stats
            path = hierarchy.format(
                tag=tag,
                network=st.network, station=st.station, channel=st.channel,
                corr_st=st.corr_start.format_fissures(),
                corr_et=st.corr_end.format_fissures())
            try:
                ds = self.create_dataset(
                    path, data=tr.data, compression=self.compression,
                    compression_opts=self.compression_opts)
                convert_header_to_hdf5(ds, st)
            except ValueError as e:
                print(e)
                warnings.warn("The dataset %s is already in file and will be \
omitted." % path, category=UserWarning)

    def get_corr_options(self) -> dict:
        try:
            sco = str(self['co'].attrs['co'])
            # Run once more through co_to_hdf5 to account for
            # correlations that have been computed with older versions

            co = co_to_hdf5(ast.literal_eval(sco))
        except KeyError:
            raise KeyError('No correlation options in file')
        return co

    def get_data(
        self, network: str, station: str, channel: str, tag: str,
        corr_start: UTCDateTime = None,
            corr_end: UTCDateTime = None) -> CorrStream:
        """
        Returns a :class:`~seismic.correlate.correlate.CorrStream` holding
        all the requested data.

        .. note::

            Wildcards are allowed for all parameters.

        :param network: network (combination), e.g., IU-YP
        :type network: str
        :param station: station (combination), e.g., HRV-BRK
        :type station: str
        :param channel: channel (combination), e.g., BZ-BR
        :type channel: str
        :param corr_start: starttime of the time windows used to computed this
            correlation, defaults to None
        :type corr_start: UTCDateTime, optional
        :param corr_end: endtime of the time windows used to computed this
            correlation, defaults to None
        :type corr_end: UTCDateTime, optional
        :return: a :class:`~seismic.correlate.correlate.CorrStream` holding
            all the requested data.
        :rtype: CorrStream
        """
        # Make sure the request is structured correctly
        sort = ['.'.join([net, stat, chan]) for net, stat, chan in zip(
                network.split('-'), station.split('-'), channel.split('-'))]
        sorted = sort.copy()
        sorted.sort()
        if sorted != sort:
            network, station, channel = ['-'.join([a, b]) for a, b in zip(
                sorted[0].split('.'), sorted[1].split('.'))]

        if isinstance(corr_start, UTCDateTime):
            corr_start = corr_start.format_fissures()
        else:
            corr_start = '*'
        if isinstance(corr_end, UTCDateTime):
            corr_end = corr_end.format_fissures()
        else:
            corr_end = '*'
        path = hierarchy.format(
            tag=tag, network=network, station=station, channel=channel,
            corr_st=corr_start, corr_et=corr_end)
        # Extremely ugly way of changing the path
        if '*' not in path and '?' not in path:
            data = np.array(self[path])
            header = read_hdf5_header(self[path])
            return CorrStream(CorrTrace(data, _header=header))
        # Now, we need to differ between the fnmatch pattern and the actually
        # accessed path
        path = path.replace('?', '*')
        pattern = path.replace('/*', '*')
        path = path.split('*')[0]
        return all_traces_recursive(self[path], CorrStream(), pattern)

    def get_available_starttimes(
        self, network: str, station: str, tag: str,
            channel: str or list = '*') -> dict:
        """
        Returns a dictionary with channel codes as keys and available
        correlation starttimes as values.

        :param network: Network code (combined code, e.g., IU-YP)
        :type network: str
        :param station: Station code (combined)
        :type station: str
        :param tag: Tag
        :type tag: str
        :param channel: Channel code (combined), wildcards allowed,
            defaults to '*'
        :type channel: str, optional
        :return: A dictionary holding the availabe starttimes for each channel
            combination
        :rtype: dict
        """
        path = hierarchy.format(
            tag=tag, network=network, station=station, channel=channel,
            corr_st='*', corr_et='*')
        out = {}
        if isinstance(channel, str):
            if '*' not in channel:
                path = '/'.join(path.split('/')[:-2])
                try:
                    out[channel] = list(self[path].keys())
                except KeyError:
                    pass
                return out
            channel = [channel]
        path = '/'.join(path.split('/')[:-3])
        for ch in channel:
            for match in fnmatch.filter(self[path].keys(), ch):
                out[match] = list(self['/'.join([path, match])].keys())
        return out

    def get_available_channels(
            self, tag: str, network: str, station: str) -> List[str]:
        """
        Returns a list of all available channels (i.e., their combination
        codes) in this file.

        :param tag: The tag that this data was save as.
        :type tag: str
        :param network: Network combination code in the form `net0-net1`
        :type network: str
        :param station: Network combination code in the form `stat0-stat1`
        :type station: str
        :return: A list of all channel combinations.
        :rtype: List[str]
        """
        path = hierarchy.format(
            tag=tag, network=network, station=station, channel='*',
            corr_st='*', corr_et='*')
        path = path.split('*')[0]
        try:
            return list(self[path].keys())
        except KeyError:
            # No channels available
            return []


class CorrelationDataBase(object):
    """
    Base class to handle the hdf5 files that contain noise correlations.
    """
    def __init__(
        self, path: str, corr_options: dict = None, mode: str = 'a',
            compression: str = 'gzip3'):
        """
        Access an hdf5 file holding correlations. The resulting file can be
        accessed using all functionalities of
        `h5py <https://www.h5py.org/>`_ (for example as a dict).

        :param path: Full path to the file
        :type path: str
        :param corr_options: The dictionary holding the parameters for the
            correlation. If set to `None`. The mode will be set to read-only
            `= 'r'`. Defaults to None.
        :type corr_options: dict or None
        :param mode: Mode to access the file. Options are: 'a' for all, 'w' for
            write, 'r+' for writing in an already existing file, or 'r' for
            read-only , defaults to 'a'.
        :type mode: str, optional
        :param compression: The compression algorithm and compression level
            that the arrays should be saved with. 'gzip3' tends to perform
            well, else you could choose 'gzipx' where x is a digit between
            1 and 9 (i.e., 9 is the highest compression) or None for fastest
            perfomance, defaults to 'gzip3'.
        :type compression: str, optional

        .. warning::

            **Access only through a context manager (see below):**

            >>> with CorrelationDataBase(myfile.h5) as cdb:
            >>>     type(cdb)  # This is a DBHandler
            <class 'seismic.db.corr_hdf5.DBHandler'>

        Example::

            >>> with CorrelationDataBase(
                        '/path/to/db/XN-XN.NEP06-NEP06.h5') as cdb:
            >>>     # find the available tags for existing db
            >>>     print(list(cdb.keys()))
            ['co', 'recombined', 'stack_86398', 'subdivision']
            >>>     # find available channels with tag subdivision
            >>>     print(cdb.get_available_channels(
            >>>         'subdivision', 'XN-XN', 'NEP06-NEP06'))
            ['HHE-HHN', 'HHE-HHZ', 'HHN-HHZ']
            >>>     # Get Data from all times, specific channel and tag
            >>>     st = cdb.get_data(
            >>>         'XN-XN', 'NEP06-NEP06', 'HHE-HHN', 'subdivision')
            >>> print(st.count())
            250
        """

        if corr_options is None and mode != 'r':
            mode = 'r'
            warnings.warn(
                'Opening Correlation Databases without providing a correlation'
                + ' options dictionary is only allowed in read only mode.\n'
                + 'Setting mode read only `r`....')
        # Create / read file
        if not path.split('.')[-1] == 'h5':
            path += '.h5'
        self.path = path
        self.mode = mode
        self.compression = compression
        self.co = corr_options

    def __enter__(self) -> DBHandler:
        self.db_handler = DBHandler(
            self.path, self.mode, self.compression, self.co)
        return self.db_handler

    def __exit__(self, exc_type, exc_value, tb) -> None or bool:
        self.db_handler._close()
        if exc_type is not None:
            return False


def all_traces_recursive(
    group: h5py._hl.group.Group, stream: CorrStream,
        pattern: str) -> CorrStream:
    """
    Recursively, appends all traces in a h5py group to the input stream.
    In addition this will check whether the data matches a certain pattern.

    :param group: group to search through
    :type group: class:`h5py._hl.group.Group`
    :param stream: Stream to append the traces to
    :type stream: CorrStream
    :param pattern: pattern for the path in the hdf5 file, see fnmatch for
        details.
    :type pattern: str
    :return: Stream with appended traces
    :rtype: CorrStream
    """
    for v in group.values():
        if isinstance(v, h5py._hl.group.Group):
            all_traces_recursive(v, stream, pattern)
        elif not fnmatch.fnmatch(v.name, pattern) and v.name not in pattern:
            continue
        else:
            stream.append(
                CorrTrace(np.array(v), _header=read_hdf5_header(v)))
    return stream


def convert_header_to_hdf5(dataset: h5py.Dataset, header: Stats):
    """
    Convert an :class:`~obspy.core.Stats` object and adds it to the provided
    hdf5 dataset.

    :param dataset: the dataset that the header should be added to
    :type dataset: h5py.Dataset
    :param header: The trace's header
    :type header: Stats
    """
    header = dict(header)
    for key in header:
        try:
            if isinstance(header[key], UTCDateTime):
                # convert time to string
                header[key] = header[key].format_fissures()
            dataset.attrs[key] = header[key]
        except TypeError:
            warnings.warn(
                'The header contains an item of type %s. Information\
            of this type cannot be written to an hdf5 file.'
                % str(type(header[key])), UserWarning)
            continue


def read_hdf5_header(dataset: h5py.Dataset) -> Stats:
    """
    Takes an hdft5 dataset as input and returns the header of the CorrTrace.

    :param dataset: The dataset to be read from
    :type dataset: h5py.Dataset
    :return: The trace's header
    :rtype: Stats
    """
    attrs = dataset.attrs
    time_keys = ['starttime', 'endtime', 'corr_start', 'corr_end']
    header = {}
    for key in attrs:
        if key in time_keys:
            try:
                header[key] = UTCDateTime(attrs[key])
            except ValueError as e:
                # temporary fix of obspy's UTCDateTime issue. SHould be removed
                # as soon as they release version 1.23
                # Actually, I can leave it. Obspy 1.23 fixes this issue, but
                # for users with older obspy it's good to be kepy
                if attrs[key][4:8] == '360T':
                    new = list(attrs[key])
                    new[6] = '1'
                    header[key] = UTCDateTime(''.join(new)) - 86400
                else:
                    raise e
        elif key == 'processing':
            header[key] = list(attrs[key])
        else:
            header[key] = attrs[key]
    return Stats(header)


def co_to_hdf5(co: dict) -> dict:
    coc = deepcopy(co)
    remk = [
        'subdir', 'read_start', 'read_end', 'read_len', 'read_inc',
        'combination_method', 'combinations', 'starttime',
        'xcombinations', 'preprocess_subdiv']
    for key in remk:
        coc.pop(key, None)
        coc['corr_args'].pop('combinations', None)
        coc['subdivision'].pop('recombine_subdivision', None)
        coc['subdivision'].pop('delete_subdivision', None)
    try:
        [coc['preProcessing'].remove(step) for step in coc['preProcessing']
            if 'stream_mask_at_utc' in step['function']]
    except KeyError:
        pass
    return coc



'''
Spatial imaging by means of a Boltzmann-equation-based sensitivity kernel
(see Obermann et al., 2013 and Paasschens, 1997)

Implementation here is just for the 2D case

:copyright:
   The SeisMIC development team (makus@gfz-potsdam.de).
:license:
    EUROPEAN UNION PUBLIC LICENCE v. 1.2
   (https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12)
:author:
   Peter Makus (makus@gfz-potsdam.de)

Created: Monday, 16th January 2023 10:53:31 am
Last Modified: Friday, 28th July 2023 03:20:46 pm
'''
from typing import Tuple, Optional, Iterator, Iterable
import warnings

import matplotlib as mpl
from matplotlib import pyplot as plt
import numpy as np
from obspy.geodetics.base import locations2degrees as loc2deg
from obspy.geodetics.base import degrees2kilometers as deg2km
from obspy import UTCDateTime

from seismic.monitor.dv import DV


def probability(
    dist: np.ndarray | float, t: np.ndarray | float, vel: float,
        mf_path: float, atol: float) -> np.ndarray | float:
    """
    Compute the probability for a wave with the given properties to pass
    through a point at distance `dist`. Following the diffusion law.

    Implementation as in Obermann et al. (2013).

    :param dist: distance to the source. Can be a matrix.
    :type dist: np.ndarray | float
    :param t: time
    :type t: float or 1D numpy array
    :param vel: velocity for homogeneous velocities
    :type vel: float
    :param mf_path: Mean free path of the wave.
    :type mf_path: float
    :return: The probability. In the same shape as ``dist``.
    :rtype: np.ndarray | float
    """
    if np.any(dist < 0):
        raise ValueError('Distances cannot be < 0.')
    # Numpy float precision can be a problem if comparing to 0
    if np.any(t < -atol/vel):
        raise ValueError('t has to be >= 0')
    # if t is an array and dist is a 2D array, create a 3D output
    # by broacasting t onto a 3D array of shape (dist.shape[0], dist.shape[1],
    # t.shape[0])
    if isinstance(t, np.ndarray):
        t = t[:, np.newaxis, np.newaxis]
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        # there will be loads of runtimeerrors because of inf and nan values
        coherent = np.nan_to_num(
            np.exp(-vel*t/mf_path)/(2*np.pi*dist),
            nan=0, posinf=1, neginf=np.nan)
        coherent *= np.isclose(dist, vel*t, atol=atol)

        a = 2*np.pi*mf_path*vel*t*np.sqrt(1 - dist**2/(vel*t)**2)
        b = (np.sqrt((vel*t)**2-dist**2) - vel*t)/mf_path
        # Sensitivity is zero where nan and 1 if infinite
        # (station coords coincide with grid coordinates)
        coda = np.nan_to_num(np.exp(b)/a)
        coda *= np.heaviside(vel*t-dist-atol, 0)
    prob = coherent + coda
    return prob


def compute_grid_dist(
        x: np.ndarray, y: np.ndarray, x0: float, y0: float) -> np.ndarray:
    """
    Compute the distance of each point on the grid x, y to x0, y0.
    (2D)

    :param x: X coordinate axis
    :type x: np.ndarray
    :param y: y coordiante axis
    :type y: np.ndarray
    :param x0: x-position of point
    :type x0: float
    :param y0: y-position of point
    :type y0: float
    :return: distance matrix in form of a 2D matrix.
    :rtype: np.ndarray
    """
    X, Y = np.meshgrid(x-x0, y-y0)
    d = np.sqrt(X**2 + Y**2)
    return d


def sensitivity_kernel(
    s1: np.ndarray, s2: np.ndarray, x: np.ndarray, y: np.ndarray, t: float,
        dt: float,  vel: float, mf_path: float) -> np.ndarray:
    """
    Computes a 2D surface-wave sensitivity kernel for an
    ambient noise cross-correlation between two station ``s1`` and ``s2``.
    The computation is based on a time-dependendent solution of the Boltzmann
    equation for a homogeneous medium (see Obermann et al., 2013 and
    Paasschens, 1997).

    Implementation as in Obermann et al. (2013). Limits for nan and inf
    will return 0, or 1 respectively.

    .. note::
        A large ``dt`` can lead to artefacts in the kernel. ``dt`` should
        always be smaller than dx/(2*vel).

    :param s1: Position of Station 1, format is [x, y].
    :type s1: np.ndarray
    :param s2: Position of Station 2, format is [x, y].
    :type s2: np.ndarray
    :param x: Monotonously increasing array holding the x-coordinates of the
        grid the kernel should be computed on.
    :type x: np.ndarray
    :param y: Monotonously increasing array holding the y-coordinates of the
        grid the kernel should be computed on.
    :type y: np.ndarray
    :param t: Time of coda that is probed. Usually set as the middle of the
        used time window
    :type t: float
    :param dt: Sampling interval that the probability will be computed in.
        Note that dt should not be too large. Otherwise, numerical artefacts
        can occur.
    :type dt: float
    :param vel: Surface wave velocity. Note that units between coordinate grid
        and velocity have to be consistent (i.e., km/s and km or m and m/s)
    :type vel: float
    :param mf_path: Mean free path of the wave (see Obermann et al. 2013 for
        details)
    :type mf_path: float
    :return: The sensitivity_kernel of the grid. Can be plotted against.\
        np.meshgrid(x, y)
    :rtype: np.ndarray
    """
    atol = vel*dt/2
    if dt <= 0:
        raise ValueError('dt has to be greater than 0.')
    if np.any(np.diff(x) != x[1]-x[0]) or np.any(np.diff(y) != y[1]-y[0]):
        raise ValueError('x and y have to be monotonously increasing arrays.')
    if x[1]-x[0] < 2*vel*dt:
        warnings.warn(
            'dt is too large for the grid spacing. This can lead to '
            'artefacts in the kernel. dt should be smaller than '
            'dx/(2*vel).')
    T = np.arange(0, t + dt, dt)
    dist_s1_s2 = np.linalg.norm(s1-s2)
    denom = probability(dist_s1_s2, t, vel, mf_path, atol=atol)
    dist_s1_x0 = compute_grid_dist(x, y, s1[0], s1[1])
    if denom == 0:
        return np.zeros_like(dist_s1_x0)
    dist_s2_x0 = compute_grid_dist(x, y, s2[0], s2[1])
    # does the direct wave have to be divided by dt?
    # This here takes way more time and RAM, I don't know why
    # nom = np.trapz(
    #     probability(
    #         dist_s1_x0, T, vel, mf_path, atol=atol) * probability(
    #         dist_s2_x0, t-T, vel, mf_path, atol=atol),
    #     dx=dt, axis=0)
    nom = np.trapz(
        [probability(
            dist_s1_x0, tt, vel, mf_path, atol=atol) * probability(
            dist_s2_x0, t-tt, vel, mf_path, atol=atol) for tt in T],
        dx=dt, axis=0)
    K = nom/denom
    return K


def data_variance(
    corr: np.ndarray | float, bandwidth: float, tw: Tuple[float, float],
        freq_c: float) -> np.ndarray | float:
    """
    Compute the variance of the velocity change data based on the coherence
    value between stretched reference trace and correlation function.

    (see Obermann et. al. 2013)

    :param corr: coherence
    :type corr: np.ndarray | float
    :param bandwidth: Bandwidth between low and highpass filter
    :type bandwidth: float
    :param tw: Time Window used for the dv/v computation. In the form
        Tuple[tw_start, tw_end]
    :type tw: Tuple[float, float]
    :param freq_c: Centre frequency.
    :type freq_c: float
    :return: Data Variance in the same shape as corr
    :rtype: np.ndarray | float
    """
    if tw[0] < 0 or tw[1] < 0:
        raise ValueError('Lapse time values have to be greater than 0s.')
    if tw[0] >= tw[1]:
        raise ValueError(
            'Lapse time values have to be in the form '
            'Tuple[tw_start, tw_end] with tw_start < tw_end.')
    if bandwidth <= 0:
        raise ValueError('The bandwidth has to be > 0 Hz.')
    if freq_c <= 0:
        raise ValueError('The centre frequeny has to be > 0 Hz.')
    T = 1/bandwidth
    p1 = np.sqrt(1-corr**2)/(2*corr)
    nom = 6*np.sqrt(np.pi/2)*T
    denom = (2*np.pi*freq_c)**2*(tw[1]**3-tw[0]**3)
    p2 = np.sqrt(nom/denom)
    return p1*p2


def compute_cm(
    scaling_factor: float, corr_len: float, std_model: float,
        dist: np.ndarray | float) -> np.ndarray | float:
    """
    Computes the model variance for the dv/v grid.

    (see Obermann et. al. 2013)

    :param scaling_factor: Scaling factor, dependent on the cell length.
        (e.g., 1*cell_length)
    :type scaling_factor: float
    :param corr_len: Length over which the parameters are related (high
        value means stronger smoothing).
    :type corr_len: float
    :param std_model: A-priori standard deviation of the model. Corresponds
        to the best agreement between 1) the stability of the model for the
        velocity variations and 2) the minimised difference between the model
        predictions (output of the forward problem) and the data.
        Can be estimated via the L-curve criterion (see Hansen 1992)
    :type std_model: float
    :param dist: Matrix containing distance between the two cells i and j.
        Will always be 0 on the diagonal where i==j.
    :type dist: np.ndarray
    :return: Returns the model variance in the same shape as dist. I.e.,
        NxN where N is the number of station combinations.
    :rtype: np.ndarray
    """
    p1 = (std_model*scaling_factor/corr_len)**2
    smooth = np.exp(-dist/corr_len)
    return p1*smooth


def geo2cart(
    lat: np.ndarray | float, lon: np.ndarray | float,
        lat0: float) -> Tuple[np.ndarray | float, np.ndarray | float]:
    """
    Convert geographic coordinates to Northing and Easting for a local grid.

    .. seealso::
        http://wiki.gis.com/wiki/index.php/Easting_and_northing

    :param lat: Latitude(s)
    :type lat: np.ndarray | float
    :param lon: Longitude(s)
    :type lon: np.ndarray | float
    :param lat0: Latitude of the lower left corner of the grid
    :type lat0: float
    :return: A tuple holding (Easting, Northing) in km and in the same shape
        as ``lat`` and ``lon``.
    :rtype: Tuple[np.ndarray, np.ndarray]
    """
    # Northing
    y = deg2km(lat)
    # easting
    x = deg2km(loc2deg(lat0, lon, lat0, 0))
    if isinstance(lon, float) or isinstance(lon, int):
        if lon < 0:
            x = -x
    else:
        x[lon < 0] *= -1
    return x, y


class DVGrid(object):
    """
    Object to handle spatially gridded DV objects.
    """
    def __init__(
        self, lat0: float, lon0: float, res: float, x: float, y: float,
            dt: float, vel: float, mf_path: float):
        """
        Initialises the grid to compute dv/v on. The grid will be in
        Cartesian coordinates with the lower left boarder (lon0, lat0)
        as reference point.

        :param lat0: Minimum latitude on the grid.
        :type lat0: float
        :param lon0: Minimum longitude on the grid
        :type lon0: float
        :param res: Resolution in km
        :type res: float
        :param x: Eastwards extent of the grid in km
        :type x: float
        :param y: Northwards extent of the grid in km
        :type y: float
        :param dt: Sampling interval for the numerical integration of the
            sensitivity kernels. Coarse spacing might cause artefacts. Fine
            spacing will result in higher computation times.
        :type dt: float
        :param vel: Wave velocity
        :type vel: float
        :param mf_path: Mean free path of a wave
        :type mf_path: float
        """
        self.lat0 = lat0
        self.lon0 = lon0

        # Compute local grid
        x0, y0 = geo2cart(lat0, lon0, lat0)
        self.xaxis = np.arange(x0, x0 + x + res, res)
        self.yaxis = np.arange(y0, y0 + y + res, res)
        self.res = res
        self.xgrid, self.ygrid = np.meshgrid(self.xaxis, self.yaxis)
        self.xf = self.xgrid.flatten()
        self.yf = self.ygrid.flatten()
        self.dist = None
        self.dt = dt
        self.vel = vel
        self.mf_path = mf_path
        # skernels is a dictionary holding the sensitivity kernels for each
        # station pair. The key is a tuple of the station coordinates
        # in cartesian coordinates and the lapse time in the coda
        # (x0, y0, x1, y1, tau)
        self.skernels = dict()

    def forward_model(
        self, dv_grid: np.ndarray, dvs: Optional[Iterable[DV]] = None,
        utc: Optional[UTCDateTime] = None,
        tw: Optional[Tuple[float, float]] = None,
        stat0: Optional[Iterable[Tuple[float, float]]] = None,
        stat1: Optional[Iterable[Tuple[float, float]]] = None) \
            -> np.ndarray:
        """
        Solves the forward problem associated to the grid computation.
        I.e., computes the velocity changes theoretically measured at each
        station for a given dv/v grid.

        :param dv_grid: dv/v grid. Must have the same shape as self.xgrid
        :type dv_grid: np.ndarray
        :param dvs: Iterator over :class:`~seismic.monitor.dv.DV` to
            extract the parameters from. If this is defined ``utc`` has to
            be defined as well. If it's None all other parameters have to be
            defined, defaults to None
        :type dvs: Optional[Iterable[DV]], optional
        :param utc: The time to extract the properties from the dv
            files from, defaults to None
        :type utc: Optional[UTCDateTime], optional
        :param tw: Lapse time Window used for the dv/v estimation,
            defaults to None
        :type tw: Optional[Tuple[float, float]], optional
        :param stat0: Coordinates of the first station of the
            cross-correlation. Given as an array with the form:
            [(latitude0, longitude0), (latitude1, longitude1), ...],
            defaults to None.
        :type stat0: Optional[Iterable[Tuple[float, float]]], optional
        :param stat1: Coordinates of the second station of the
            cross-correlation. Given as an array with the form:
            [(latitude0, longitude0), (latitude1, longitude1), ...],
            defaults to None., defaults to None
        :type stat1: Optional[Iterable[Tuple[float, float]]], optional
        :raises TypeError: If too few variables are defined
        :return: A vector with one dv/v value for each station combination
            in the stat0-stat1 vectors.
        :rtype: np.ndarray
        """
        if dv_grid.shape != self.xgrid.shape:
            raise ValueError(
                'Modelled dvgrid must have same shape as self.')
        if stat0 is not None and stat1 is not None and tw is not None:
            slat0, slon0 = np.array(list(zip(*stat0)))
            slat1, slon1 = np.array(list(zip(*stat1)))
            twe = None
        elif dvs is not None and utc is not None:
            _, _, slat0, slon0, slat1, slon1, twe, _, _\
                = self._extract_info_dvs(dvs, utc)
        else:
            raise TypeError(
                'Define either dv and utc or all other arguments.'
            )

        tw = tw or twe
        G = self._compute_sensitivity_kernels(
            slat0, slon0, slat1, slon1, (tw[0]+tw[1])/2)
        return np.dot(G, dv_grid.flatten())

    def compute_dv_grid(
        self, dvs: Iterator[DV], utc: UTCDateTime, scaling_factor: float,
        corr_len: float, std_model: float,
        tw: Optional[Tuple[float, float]] = None,
        freq0: Optional[float] = None, freq1: Optional[float] = None,
            compute_resolution: bool = False) -> np.ndarray:
        """
        Perform a linear least-squares inversion of station specific velocity
        changes (dv/v) at time ``t``.

        The algorithm will perform the following steps:
        1. Computes the sensitivity kernels for each station pair.
        2. Computes model and data variance.
        3. Inverts the velocity changes onto a flattened grid.

        .. note::
            All computations are performed on a flattened grid, so that the
            grid is actually just one-dimensional. The return value is
            reshaped onto the shape of self.xgrid.

        .. note::
            The returned value and the value in self.vel_change is strictly
            speaking not a velocity change, but an epsilon value. That
            corresponds to -dv/v


        :param dvs: The :class:``~seismic.monitor.dv.DV`` objects to use.
            Note that the input can be an iterator, so that not actually all
            dvs have to be loaded into RAM at the same time.
        :type dvs: Iterator[DV]
        :param utc: Time at which the velocity changes should be inverted.
            Has to be covered by the ``DV`` objects.
        :type utc: UTCDateTime
        :param scaling_factor: Scaling factor, dependent on the cell length.
            (e.g., 1*cell_length)
        :type scaling_factor: float
        :param corr_len: Length over which the parameters are related (high
            value means stronger smoothing).
        :type corr_len: float
        :param std_model: A-priori standard deviation of the model. Corresponds
            to the best agreement between 1) the stability of the model for the
            velocity variations and 2) the minimised difference between the
            model predictions (output of the forward problem) and the data.
            Can be estimated via the L-curve criterion (see Hansen 1992)
        :type std_model: float
        :param tw: Time window used to evaluate the velocity changes.
            Given in the form Tuple[tw_start, tw_end]. Not required if
            the used dvs have a ``dv_processing`` attribute.
        :type tw: Tuple[float, float], optional
        :param freq0: high-pass frequency of the used bandpass filter.
            Not required if the used dvs have a ``dv_processing`` attribute.
        :type freq0: float, optional
        :param freq1: low-pass frequency of the used bandpass filter.
            Not required  the used dvs have a ``dv_processing`` attribute.
        :type freq1: float, optional
        :param compute_resolution: Compute the resolution matrix?
            Will be saved in self.resolution
        :type compute_resolution: bool, defaults to False
        :raises ValueError: For times that are not covered by the dvs
            time-series.
        :return: The dv-grid (with corresponding coordinates in
            self.xgrid and self.ygrid or self.xaxis and self.yaxis,
            respectively). Note this is actually an epsilon grid (meaning
            it represents the stretch value i.e., -dv/v)
        :rtype: np.ndarray
        """
        vals, corrs, slat0, slon0, slat1, slon1, twe, freq0e, freq1e\
            = self._extract_info_dvs(dvs, utc)
        tw = tw or twe
        freq0 = freq0 or freq0e
        freq1 = freq1 or freq1e
        if None in [freq0, freq1, tw]:
            raise AttributeError(
                'The arguments freq0, freq1, and tw have to be defined if '
                'dv does not have the attribute "dv_processing".'
            )
        skernels = self._compute_sensitivity_kernels(
            slat0, slon0, slat1, slon1, (tw[0]+tw[1])/2)
        dist = self._compute_dist_matrix()
        # Model variance, smoothed diagonal matrix
        cm = compute_cm(scaling_factor, corr_len, std_model, dist)
        # Data variance, diagonal matrix
        cd = self._compute_cd(skernels, freq0, freq1, tw, corrs)

        # Linear Least-Squares Inversion
        a = np.dot(cm, skernels.T)
        b = np.linalg.inv(np.dot(np.dot(skernels, cm), skernels.T) + cd)
        c = np.array(vals)
        m = np.dot(np.dot(a, b), c)
        # m is the flattened array, reshape onto grid
        self.vel_change = m.reshape(self.xgrid.shape)
        if compute_resolution:
            self.resolution = self._compute_resolution(skernels, a, b)
        return self.vel_change

    def compute_resolution(
        self, dvs: Iterable[DV], utc: UTCDateTime, scaling_factor: float,
        corr_len: float, std_model: float,
        tw: Optional[Tuple[float, float]] = None,
        freq0: Optional[float] = None, freq1: Optional[float] = None) \
            -> np.ndarray:
        """
        Compute the model resolution of the dv/v model.

        .. seealso::
            Described in the supplement to Obermann (2013)

        .. note::
            Note that it's significantly faster to perform a joint inversion
            or a gridded dv/v and resolution rather than performing them
            separately.

        :param dvs: The :class:``~seismic.monitor.dv.DV`` objects to use.
            Note that the input can be an iterator, so that not actually all
            dvs have to be loaded into RAM at the same time.
        :type dvs: Iterator[DV]
        :param utc: Time at which the velocity changes should be inverted.
            Has to be covered by the ``DV`` objects.
        :type utc: UTCDateTime
        :param scaling_factor: Scaling factor, dependent on the cell length.
            (e.g., 1*cell_length)
        :type scaling_factor: float
        :param corr_len: Length over which the parameters are related (high
            value means stronger smoothing).
        :type corr_len: float
        :param std_model: A-priori standard deviation of the model. Corresponds
            to the best agreement between 1) the stability of the model for the
            velocity variations and 2) the minimised difference between the
            model predictions (output of the forward problem) and the data.
            Can be estimated via the L-curve criterion (see Hansen 1992)
        :type std_model: float
        :param tw: Time window used to evaluate the velocity changes.
            Given in the form Tuple[tw_start, tw_end]. Not required if dv
            has attribute ``dv_processing``.
        :type tw: Tuple[float, float], optional
        :param freq0: high-pass frequency of the used bandpass filter.
            Not required if dv has attribute ``dv_processing``.
        :type freq0: float, optional
        :param freq1: low-pass frequency of the used bandpass filter.
            Not required if dv has attribute ``dv_processing``.
        :type freq1: float, optional
        :raises ValueError: For times that are not covered by the dvs
            time-series.
        :return: A gridded resolution matrix.
        :rtype: np.ndarray
        """
        cm = compute_cm(
            scaling_factor, corr_len, std_model,
            self.dist or self._compute_dist_matrix())
        _, corrs, slat0, slon0, slat1, slon1, twe, freq0e, freq1e\
            = self._extract_info_dvs(dvs, utc)
        freq0 = freq0 or freq0e
        freq1 = freq1 or freq1e
        tw = tw or twe
        if None in [freq0, freq1, tw]:
            raise AttributeError(
                'The arguments freq0, freq1, and tw have to be defined if '
                'dv does not have the attribute "dv_processing".'
            )
        skernels = self._compute_sensitivity_kernels(
            slat0, slon0, slat1, slon1, (tw[0]+tw[1])/2)
        cd = self._compute_cd(skernels, freq0, freq1, tw, corrs)

        # The actual inversion for the resolution
        a = np.dot(cm, skernels.T)
        b = np.linalg.inv(np.dot(np.dot(skernels, cm), skernels.T) + cd)
        self.resolution = self._compute_resolution(skernels, a, b)
        return self.resolution

    def _compute_resolution(
        self, skernels: np.ndarray, a: np.ndarray,
            b: np.ndarray) -> np.ndarray:
        """
        The actual computation of the model resolution.
        """
        res = np.sum(np.dot(np.dot(a, b), skernels), axis=-1)
        self.resolution = res.reshape(self.xgrid.shape)
        return self.resolution

    def _extract_info_dvs(
        self, dvs: Iterable[DV], utc: UTCDateTime) -> Tuple[
            np.ndarray, np.ndarray, np.ndarray, np.ndarray,
            Tuple[float, float], float, float]:
        """
        Extract the reqired infos from a iterator over dv.

        :param dvs: station-specific dvs to be used for the inversion
        :type dvs: Iterable[DV]
        :param utc: Time to perform the inversion for
        :type utc: UTCDateTime
        :raises ValueError: if ``utc`` is not covered by dvs
        :return: dv/v estimates, coherency values, latituds_stat1,
            longitudes_stat1, latitudes_stat2, longitudes_stat2.
            All specific for time ``utc``
        :rtype: Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
        """
        vals, corrs, slat0, slon0, slat1, slon1 = [], [], [], [], [], []
        for dv in dvs:
            if utc < dv.stats.corr_start[0] or utc > dv.stats.corr_end[-1]:
                raise IndexError(
                    f'Time {utc} is outside of the dv time-series'
                )
            ii = np.argmin(abs(np.array(dv.stats.corr_start)-utc))
            val = dv.value[ii]
            corr = dv.corr[ii]
            if np.isnan(val) or np.isnan(corr):
                # No value
                continue
            vals.append(val)
            corrs.append(corr)
            slat0.append(dv.stats.stla)
            slon0.append(dv.stats.stlo)
            slat1.append(dv.stats.evla)
            slon1.append(dv.stats.evlo)
            processing = dv.dv_processing
        if not len(vals):
            raise IndexError(
                f'None of the given dvs have available data at {utc}.')
        slat0 = np.array(slat0)
        slon0 = np.array(slon0)
        self._add_stations(slat0, slon0)
        slat1 = np.array(slat1)
        slon1 = np.array(slon1)
        self._add_stations(slat1, slon1)
        vals = np.array(vals)
        corrs = np.array(corrs)
        if processing is not None:
            tw = (
                processing['tw_start'],
                processing['tw_start'] + processing['tw_len'])
            freq0 = processing['freq_min']
            freq1 = processing['freq_max']
        else:
            tw = None
            freq0 = None
            freq1 = None
        return vals, corrs, slat0, slon0, slat1, slon1, tw, freq0, freq1

    def _add_stations(self, lats: Iterable[float], lons: Iterable[float]):
        """
        Add Coordinates of the stations to self (coordinates are converted
        from geographic coordinates to northing and easting).

        :param lats: Station Latitudes
        :type lats: Iterable[float]
        :param lons: Station Longitudes
        :type lons: Iterable[float]
        """
        x, y = geo2cart(np.array(lats), np.array(lons), self.lat0)
        if hasattr(self, 'statx'):
            self.statx = np.hstack((self.statx, x))
            self.staty = np.hstack((self.staty, y))
            self.statlon = np.hstack((self.statlon, lons))
            self.statlat = np.hstack((self.statlat, lats))
        else:
            self.statx = np.array(x)
            self.staty = np.array(y)
            self.statlon = np.array(lons)
            self.statlat = np.array(lats)
        # make sure there are no duplicates
        coords = [(x, y) for x, y in zip(self.statx, self.staty)]
        coords = list(set(coords))
        self.statx, self.staty = zip(*coords)
        self.statx = np.array(self.statx)
        self.staty = np.array(self.staty)
        # no duplicates for lat/lon
        coords = [(lat, lon) for lat, lon in zip(self.statlat, self.statlon)]
        coords = list(set(coords))
        self.statlat, self.statlon = zip(*coords)
        self.statlat = np.array(self.statlat)
        self.statlon = np.array(self.statlon)

    def _compute_cd(
        self, skernels: np.ndarray, freq0: float, freq1: float,
            tw: Tuple[float, float], corrs: np.ndarray) -> np.ndarray:
        """
        Compute the Variance of the data (in matrix form).

        :param skernels: Sensitivity kernel matrix
        :type skernels: np.ndarray
        :param freq0: highpass frequency [Hz]
        :type freq0: float
        :param freq1: lowpass frequency [Hz]
        :type freq1: float
        :param tw: lapse time window used to determine dv/v, in the form
            Tuple[tw_start, tw_end]
        :type tw: Tuple[float, float]
        :param corrs: Coherency values.
        :type corrs: np.ndarray
        :return: A diagonal matrix describing the data variance.
        :rtype: np.ndarray
        """
        return np.eye(len(skernels))*data_variance(
            np.array(corrs), freq1-freq0, tw, (freq1+freq0)/2)**2

    def _compute_dist_matrix(self) -> np.ndarray:
        """
        Computes the distance from each cell to one another.

        :return: 2D matrix holding distances between cells i and j.
        :rtype: np.ndarray
        """
        self.dist = np.zeros((len(self.xf), len(self.xf)))
        for ii, (xx, yy) in enumerate(zip(self.xf, self.yf)):
            self.dist[ii, :] = np.sqrt((self.xf-xx)**2+(self.yf-yy)**2)
        return self.dist

    def _find_coord(
        self, lat: float | np.ndarray,
            lon: float | np.ndarray) -> int | np.ndarray:
        """
        Finds a coordinate given in lat, lon on the Cartesian grid and
        return its index

        ..note:: The grid is flattened, so only one index i is returned
            per lat/lon combination

        :param lat: Latitude(s) of the points to determine
        :type lat: float | np.ndarray
        :param lon: Longitude(s) of the points to determine
        :type lon: float | np.ndarray
        :return: The indices of the points. Will have the same length
            as lat/lon (note that either of the two can be a floating point
            number).
        :rtype: int | np.ndarray
        """
        # Find northing and easting of coordinates
        x, y = geo2cart(lat, lon, self.lat0)

        if isinstance(x, float) or isinstance(x, int):
            if np.all(abs(self.xf-x) > self.res) \
                    or np.all(abs(self.yf-y > self.res)):
                raise ValueError(
                    'The point is outside the coordinate grid'
                )
            return np.argmin(abs(self.xf-x)+abs(self.yf-y))
        if np.all(np.array([abs(self.xf-xx) for xx in x]) > self.res)\
                or np.all(np.array([abs(self.yf-yy) for yy in y]) > self.res):
            raise ValueError(
                'The point is outside the coordinate grid'
            )
        ii = np.array(
            [np.argmin(
                abs(self.xf-xx)+abs(self.yf-yy)) for xx, yy in zip(x, y)])
        return ii

    def _compute_sensitivity_kernels(
        self, slat0: np.ndarray, slon0: np.ndarray, slat1: np.ndarray,
            slon1: np.ndarray, t: float) -> np.ndarray:
        """
        Computes the sensitivity kernels of the station combinations station0
        and station1.

        .. note::
            Station0 and station1 may be identical. Then, the output
            corresponds to the sensitivity kernel of an auto or self
            correlation.

        .. note::
            Handles computations for several dv objects, so the station
            coordinates can be 1D arrays.

        .. warning::
            Low values for ``dt`` can produce artefacts. Very high values
            lead to a drastically increased computational cost.

        :param slat0: Latitude of the first station.
        :type slat0: np.ndarray
        :param slon0: Longitude of the first station
        :type slon0: np.ndarray
        :param slat1: Latitude of the second station.
        :type slat1: np.ndarray
        :param slon1: Longitude of the second station.
        :type slon1: np.ndarray
        :param t: Midpoint of the time lapse window used to evaluate dv/v.
        :type t: float
        :return: Weighted matrix holding the sensitivity kernel for station i
            on grid j
        :rtype: np.ndarray
        """
        x0, y0 = geo2cart(slat0, slon0, self.lat0)
        x1, y1 = geo2cart(slat1, slon1, self.lat0)
        skernels = []

        for xx0, yy0, xx1, yy1 in zip(x0, y0, x1, y1):
            # Check if skernel was already computed in a previous iteration
            if (xx0, yy0, xx1, yy1, t) in self.skernels:
                skernels.append(self.skernels[(xx0, yy0, xx1, yy1, t)])
                continue
            # Station coordinates in cartesian
            s1 = np.array([xx0, yy0])
            s2 = np.array([xx1, yy1])
            sk = sensitivity_kernel(
                s1, s2, self.xaxis, self.yaxis, t, self.dt, self.vel,
                self.mf_path)
            skernels.append(self.res**2/t * sk.flatten())
            self.skernels[(xx0, yy0, xx1, yy1, t)] = skernels[-1]
        return np.array(skernels)

    def plot(
        self, plot_stations: bool = True, cmap: str = 'seismic_r',
        ax: Optional[mpl.axis.Axis] = None, *args,
            **kwargs) -> mpl.axis.Axis:
        """
        Simple heatplot of the dv/v grid in Cartesian coordinates.

        :param plot_stations: plot station coordinates, defaults to True
        :type plot_stations: bool, optional
        :param cmap: Colormap to use, defaults to 'seismic'
        :type cmap: str, optional
        :param ax: Axis to plot into, defaults to None
        :type ax: Optional[mpl.axis.Axis], optional
        :return: Returns the axis handle
        :rtype: mpl.axis.Axis
        """
        if not ax:
            plt.figure()
            ax = plt.gca()
        norm = mpl.colors.TwoSlopeNorm(vcenter=0)
        map = ax.imshow(
            -100*np.flipud(self.vel_change), interpolation='none',
            extent=[
                self.xaxis.min(), self.xaxis.max(), self.yaxis.min(),
                self.yaxis.max()],
            cmap=cmap, zorder=0, norm=norm, *args, **kwargs)
        if plot_stations:
            ax.scatter(
                self.statx, self.staty, marker='v', c='k', edgecolors='white',
                s=30)
        plt.colorbar(
            map, label=r'$\frac{dv}{v}$ [%]', orientation='horizontal')
        plt.xlabel('Easting [km]')
        plt.ylabel('Northing [km]')
        return ax



'''
:copyright:
    The SeisMIC development team (makus@gfz-potsdam.de).
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
:author:
    Peter Makus (makus@gfz-potsdam.de)

Created: Thursday, 8th July 2021 10:37:21 am
Last Modified: Friday, 10th February 2023 04:09:21 pm
'''
from collections.abc import Iterable

from obspy.core.utcdatetime import UTCDateTime
from obspy.core.trace import Stats
from scipy.io import loadmat
from seismic.correlate.stats import CorrStats

from seismic.correlate.stream import CorrTrace, CorrBulk


def flatten(x):
    """ Return the flattened version of the input array x

    This funciton works with all :class:`~collections.Iterable` that can be
    nested in an irregular fashion.

    :type x: :class:`~collections.Iterable`
    :param: Iterable to be flattened

    :rtype: list
    :return: x as a flattened list

    .. rubric: Example
        >>> L=[[[1, 2, 3], [4, 5]], 6]
        >>> flatten(L)
        [1, 2, 3, 4, 5, 6]
    """
    result = []
    for el in x:
        if isinstance(el, Iterable) and not isinstance(el, str):
            result.extend(flatten(el))
        else:
            result.append(el)
    return result


def flatten_recarray(x):
    """ Flatten a recarray
    """
    flat_dict = {}
    if hasattr(x, 'dtype'):
        if hasattr(x.dtype, 'names'):
            if x.dtype.names is not None:
                for name in x.dtype.names:
                    field = flatten(x[name])
                    if field != []:
                        flat_dict.update({name: field[0]})
                    else:
                        flat_dict.update({name: field})
                return flat_dict
    return x


zerotime = UTCDateTime(1971, 1, 1)


def mat_to_corrtrace(mat: dict) -> CorrTrace:
    """
    Reads in a correlation dictionary created with the old (i.e., python2)
    miic and translates it into a :class:`~seismic.correlate.stream.CorrTrace`
    object.

    :param mat: dictionary as produced by :func:`scipy.io.loadmat`
    :type mat: dict
    :return: A Correlation Trace
    :rtype: CorrTrace
    """
    for key in mat:
        if mat[key] is not None:
            flat_var = flatten_recarray(mat[key])
            mat.update({key: flat_var})
    start_lag = UTCDateTime(mat['stats']['starttime']) - zerotime
    end_lag = UTCDateTime(mat['stats']['endtime']) - zerotime
    if not isinstance(mat['stats_tr1']['location'], str):
        mat['stats_tr1'].update({'location': '-'})
    if not isinstance(mat['stats_tr2']['location'], str):
        mat['stats_tr2'].update({'location': '-'})
    ctr = CorrTrace(
        mat['corr_trace'].flatten(), Stats(mat['stats_tr1']),
        Stats(mat['stats_tr2']), start_lag=start_lag, end_lag=end_lag)
    return ctr


def corrmat_to_corrbulk(path: str) -> CorrBulk:
    """
    Reads in a correlation .mat file created with the old (i.e., python2)
    miic and translates it into a :class:`~seismic.correlate.stream.CorrTrace`
    object.

    :param path: Path to file
    :type path: str
    :return: A Correlation Bulk object
    :rtype: CorrBulk

    :warning: This reads both stacks and subdivisions into the same object.
    """
    mat = loadmat(path)
    for key in mat:
        if mat[key] is not None:
            flat_var = flatten_recarray(mat[key])
            mat.update({key: flat_var})
    stats = CorrStats()
    start_lag = UTCDateTime(mat['stats']['starttime']) - zerotime
    data = mat.pop('corr_data')
    corr_start = [UTCDateTime(t) for t in mat['time']]
    mat['stats'].pop('starttime')
    stats.update(mat['stats'])
    stats.update({'start_lag': start_lag, 'corr_start': corr_start})
    return CorrBulk(data, stats)



'''
Manage objects holding correlations.

:copyright:
    The SeisMIC development team (makus@gfz-potsdam.de).
:license:
    EUROPEAN UNION PUBLIC LICENCE v. 1.2
   (https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12)
:author:
   Peter Makus (makus@gfz-potsdam.de)

Created: Tuesday, 20th April 2021 04:19:35 pm
Last Modified: Thursday, 24th August 2023 11:06:37 am
'''
from typing import Iterator, List, Tuple, Optional
from copy import deepcopy
import warnings
from matplotlib import pyplot as plt


import numpy as np
import numpy.typing as npt
from obspy import Stream, Trace, Inventory, UTCDateTime
from obspy.core import Stats

from seismic.utils import miic_utils as m3ut
from seismic.plot.plot_correlation import plot_cst, plot_ctr, plot_corr_bulk
import seismic.monitor.post_corr_process as pcp
from seismic.monitor.stretch_mod import wfc_multi_reftr
from seismic.monitor.dv import DV
from seismic.monitor.wfc import WFC
from seismic.correlate.stats import CorrStats
from seismic.monitor.trim import corr_mat_trim


class CorrBulk(object):
    """
    An object for faster computations on several correlations. The input
    correlation contain data from only one Station-Channel pair.
    """
    def __init__(
        self, A: np.ndarray, stats: CorrStats = None,
            statlist: List[CorrStats] = None):
        """
        An object for faster computations on several correlations. The input
        correlation contain data from only one Station-Channel pair.

        :param A: the Correlation Matrix. Each line corresponds to one
            correlation start. The column distance is lag time
        :type A: np.ndarray
        :param stats: CorrStats file holding the header for this object,
            only relevant if you are reloading this file. Otherwise,
            use statlist as input to create this header.
            defaults to None
        :type stats: CorrStats, optional
        :param statlist: Header of each CorrTrace used to create this object,
            defaults to None
        :type statlist: List[CorrStats], optional
        """
        self.data = A
        if stats:
            self.stats = stats
        elif statlist:
            self.stats = convert_statlist_to_bulk_stats(statlist)
        else:
            self.stats = CorrStats()
            self.stats['ntrcs'], self.stats['npts'] = A.shape
        self.stats['processing_bulk'] = []
        self.ref_trc = None

    def plot(self, **kwargs):
        """
        Plot the correlation traces contained.

        .. seealso::
            See :func:`~seismic.plot.plot_correlation.plot_corr_bulk` for
            accepted parameters.
        """
        ax = plot_corr_bulk(self, **kwargs)
        return ax

    def normalize(
        self, starttime: float = None, endtime: float = None,
            normtype: str = 'energy'):
        """
        Correct amplitude variations with time.

        Measure the maximum of the absolute value of the correlation matrix in
        a specified lapse time window and normalize the correlation traces by
        this values. A coherent phase in the respective lapse time window will
        have constant ampitude afterwards..

        :type starttime: float
        :param starttime: Beginning of time window in seconds with respect to
            the zero position.
        :type endtime: float
        :param endtime: end time window in seconds with respect to the zero
            position.
        :type normtype: string
        :param normtype: one of the following 'energy', 'max', 'absmax',
            'abssum' to decide about the way to calculate the normalization.
        :rtype: CorrBulk
        :return: Same object as in self, but normalised.

        ..note:: This action is performed **in-place**. If you would like to
            keep the original data use
            :func:`~seismic.correlate.stream.CorrelationBulk.copy()`.
        """
        self.data = pcp.corr_mat_normalize(
            self.data, self.stats, starttime, endtime, normtype)
        proc_str = f'normalize; normtype: {normtype}'
        if starttime is not None and endtime is not None:
            proc_str += f', starttime: {starttime}, endtime: {endtime}'
        self.stats.processing_bulk += [proc_str]
        return self

    def clip(self, thres: float, axis=1):
        """
        Clip the correlation data's upper and lower bounds to a multiple of its
        standard deviation. `thres` determines the factor and `axis` the axis
        the std should be computed over

        :param thres: factor of the standard deviation to clip by
        :type thres: float
        :param axis: Axis to compute the std over and, subsequently clip over.
            Can be None, if you wish to compute floating point rather than a
            vector. Then, the array will be clipped evenly.
        :type axis: int

        ..note:: This action is performed **in-place**. If you would like to
                keep the original data use
                :func:`~seismic.correlate.stream.CorrelationBulk.copy()`.
        """
        self.data = pcp.corr_mat_clip(self.data, thres, axis)
        proc_str = f'Clipped; threshold: {thres}*std, axis={axis}'
        self.stats.processing_bulk += [proc_str]
        return self

    def copy(self):
        """
        Returns a copy of self

        :return: A copy of self
        """
        return deepcopy(self)

    def correct_decay(self):
        """
        Correct for the amplitude decay in a correlation matrix.

        Due to attenuation and geometrical spreading the amplitude of the
        correlations decays with increasing lapse time. This decay is corrected
        by dividing the correlation functions by an exponential function that
        models the decay.

        :return: Self but with data corrected for amplitude decay

        ..note:: This action is performed **in-place**. If you would like to
            keep the original data use
            :func:`~seismic.correlate.stream.CorrelationBulk.copy()`.
        """
        self.data = pcp.corr_mat_correct_decay(self.data, self.stats)
        self.stats.processing_bulk += ['Corrected for Amplitude Decay']
        return self

    def correct_stretch(self, dv: DV):
        """
        Correct stretching of correlation matrix

        In the case of a homogeneous subsurface velocity change the correlation
        traces are stretched or compressed. This stretching can be measured
        with `self.stretch`. The resulting `DV` object can be passed to
        this function to remove the stretching from the correlation matrix.

        :param dv: Velocity Change object
        :type dv: DV

        ..note:: This action is performed **in-place**. If you would like to
            keep the original data use
            :func:`~seismic.correlate.stream.CorrelationBulk.copy()`.
        """
        if dv.value_type != 'stretch':
            raise ValueError('DV object does not hold any stretch values.')
        self.data, self.stats = pcp.apply_stretch(
            self.data, self.stats, -1.*dv.value)
        self.stats.processing_bulk += ['Applied time stretch']
        return self

    def correct_shift(self, dt: DV):
        """
        Correct a shift of the traces.

        If time shifts in the (correlation) traces occur due to clock drifts
        or time offsets in active measurements these can be measured with
        :func:`~seismic.correlate.stream.CorrelationBulk.measure_shift`. If
        the resulting time shift is passed to this function the shift is
        corrected for, such that if the measurement is done again no shift will
        be detected.
        """
        self.data = pcp.apply_shift(
            data=self.data, stats=self.stats,
            shifts=-1.*dt.value)
        self.stats.processing_bulk += ['Corrected for time shift']
        return self

    def create_corr_stream(self, ind: List[int] = None):
        """
        Creates a :class:`~seismic.correlate.stream.CorrStream` object from
        the current :class:`~seismic.correlate.stream.CorrBulk` object. This
        can be useful if you want to save your postprocessed data in a hdf5
        file again.

        :param ind: Indices to extract. If None, all will be used.
            Defaults to None.
        :type ind: Iterable[ind], optional
        :return: Correlation Stream holding the same data
        :rtype: :class:`~seismic.correlate.stream.CorrStream`
        """
        cst = CorrStream()
        mutables = ['corr_start', 'corr_end', 'location', 'channel']
        if ind is None:
            for ii, li in enumerate(self.data):
                stats = deepcopy(self.stats)
                for k in mutables:
                    if not isinstance(stats[k], list):
                        continue
                    stats[k] = self.stats[k][ii]
                ctr = CorrTrace(li, _header=stats)
                cst.append(ctr)
        else:
            for ii in ind:
                stats = deepcopy(self.stats)
                for k in mutables:
                    if not isinstance(stats[k], list):
                        continue
                    stats[k] = self.stats[k][ii]
                ctr = CorrTrace(self.data[ii], _header=stats)
                cst.append(ctr)
        return cst

    def envelope(self):
        """
        Calculate the envelope of a correlation matrix.

        The correlation data of the correlation matrix are replaced by their
        Hilbert envelopes.


        :return: self with the envelope in data

        ..note:: This action is performed **in-place**. If you would like to
            keep the original data use
            :func:`~seismic.correlate.stream.CorrelationBulk.copy()`.
        """
        self.data = pcp.corr_mat_envelope(self.data)
        self.stats.processing_bulk += ['Computed Envelope']
        return self

    def filter(self, freqs: Tuple[float, float], order: int = 3):
        """
        Filters the correlation matrix in the frequency band specified in
        freqs using a zero phase filter of twice the order given in order.

        :type freqs: Tuple
        :param freqs: lower and upper limits of the pass band in Hertz
        :type order: int
        :param order: half the order of the Butterworth filter

        :return: self

        ..note:: This action is performed **in-place**. If you would like to
            keep the original data use
            :func:`~seismic.correlate.stream.CorrelationBulk.copy()`.
        """
        self.data = pcp.corr_mat_filter(self.data, self.stats, freqs, order)
        proc = [f'filter; freqs: {freqs}, order: {order}']
        self.stats.processing_bulk += proc
        return self

    def extract_trace(
        self, method: str = 'mean',
            percentile: float = 50.) -> np.ndarray:
        """
        Extract a representative trace from a correlation matrix.

        Extract a correlation trace from the that best represents the
        correlation matrix. ``Method`` decides about method to extract the
        trace. The following possibilities are available

        * ``mean`` averages all traces in the matrix
        * ``median`` takes the median of all traces in the matrix
        * ``norm_mean`` averages the traces normalized after normalizing for
            maxima
        * ``similarity_percentile`` averages the ``percentile`` % of traces
            that best correlate with the mean of all traces. This will exclude
            abnormal traces. ``percentile`` = 50 will return an average of
            traces with correlation (with mean trace) above the median.

        :type method: string
        :param method: method to extract the trace
        :type percentile: float
        :param percentile: only used for method=='similarity_percentile'

        :rtype: np.ndarray
        :return: extracted trace

        ..note:: the extracted trace will also be saved in self.ref_trc
        """
        outdata = pcp.corr_mat_extract_trace(
            self.data, self.stats, method, percentile)
        self.ref_trc = outdata
        return outdata

    def extract_multi_trace(
        self, win_inc: int or List[int], method: str = 'mean',
            percentile: float = 50.) -> List[np.ndarray]:
        """
        Extract several representative traces from a correlation matrix.
        (one per time window `win_inc` with the length = 2*win_inc). That is,
        the reference will be extracted with half increment overlap in each
        direction.

        Extract a correlation trace from the one that best represents the
        correlation matrix. ``Method`` decides about method to extract the
        trace. The following possibilities are available

        * ``mean`` averages all traces in the matrix
        * ``median`` extract the median of all traces in the matrix
        * ``norm_mean`` averages the traces normalized after normalizing for
            maxima
        * ``similarity_percentile`` averages the ``percentile`` % of traces
            that best correlate with the mean of all traces. This will exclude
            abnormal traces. ``percentile`` = 50 will return an average of
            traces with correlation (with mean trace) above the median.

        :param win_inc: Increment between each window that a Trace should be
            extracted from. Given in days. Can be a list/np.ndarray or an int
            (constant increment). The windows' length will be twice the
            increment. **If set to 0, only one trace for the whole time will
            be used.**
        :type win_inc: int or List[int] (number of days)
        :type method: string
        :param method: method to extract the trace
        :type percentile: float
        :param percentile: only used for method=='similarity_percentile'
        :rtype: np.ndarray
        :return: extracted trace

        ..note:: the extracted traces will also be saved in self.ref_trc
        """
        ref_trcs = []
        if isinstance(win_inc, list):
            win_inc = np.array(win_inc)
        elif win_inc == 0:
            return self.extract_trace(method, percentile)

        inc_s = win_inc*24*3600
        start = min(self.stats.corr_start)
        if isinstance(inc_s, np.ndarray):
            for inc in inc_s:
                end = start + inc
                ii = self._find_slice_index(start-inc/2, end+inc/2, True)
                start = start + inc
                ref_trcs.append(pcp.corr_mat_extract_trace(
                    self.data[ii, :], self.stats, method, percentile))
        else:
            while start < max(self.stats.corr_end):
                end = start + inc_s
                ii = self._find_slice_index(start, end, True)
                start = end
                # stats don't need to be altered for this function
                ref_trcs.append(pcp.corr_mat_extract_trace(
                    self.data[ii, :], self.stats, method, percentile))
        ref_trcs = np.array(ref_trcs)
        self.ref_trc = ref_trcs
        return ref_trcs

    def find_clock_shift(
        self, ref_trc: np.ndarray = None, tw: List[np.ndarray] = None,
        shift_range: int = 10, shift_steps: int = 101,
            sides: str = 'both', return_sim_mat: bool = False) -> DV:
        """
        Compute the shift of correlations as they can occur due to a clock
        drift.

        :param ref_trc: Reference trace to use for the computation,
            defaults to None. Will extract a single trace if = None.
        :type ref_trc: np.ndarray, optional
        :param tw: Lapse Time window to use for the computation,
            defaults to None
        :type tw: List[np.ndarray], optional
        :param shift_range: Maximum shift value to test
            (in n samples not seconds!). Defaults to 10.
        :type shift_range: int, optional
        :param shift_steps: Number of shift steps, defaults to 101
        :type shift_steps: int, optional
        :param sides: Which sides to use. Can be 'right',
            or 'both'. Defaults to 'both'
        :type sides: str, optional
        :param return_sim_mat: Return the similarity matrix, defaults to False
        :type return_sim_mat: bool, optional
        :return: The shift as :class:`~seismic.monitor.dv.DV` object.
        :rtype: DV
        """
        if ref_trc is None:
            ref_trc = self.ref_trc
        dv_dict = pcp.corr_mat_shift(
            self.data, self.stats, ref_trc, tw, shift_range, shift_steps,
            sides, return_sim_mat)
        if not return_sim_mat:
            dv_dict['sim_mat'] = np.array([])
        return DV(**dv_dict)

    def measure_shift(
        self, ref_trc: Optional[np.ndarray] = None,
        tw: Optional[List[float]] = None,
        shift_range: float = 10, shift_steps: int = 101, sides: str = 'both',
            return_sim_mat: bool = False) -> DV:
        """
        Time shift estimate through shifting and comparison.

        This function estimates shifting of the time axis of traces as it can
        occur if the clocks of digitizers drift.

        Time shifts are estimated comparing each trace (e.g. correlation
        function stored in the ``corr_data`` matrix (one for each row) with
        shifted versions  of reference trace stored in ``ref_trc``. The range
        of shifting to be tested is given in ``shift_range`` in seconds.
        It is used in a symmetric way from -``shift_range``
        to +``shift_range``. Shifting ist
        tested ``shift_steps`` times. ``shift_steps`` should be an odd number
        to test zero shifting. The best match (shifting amount and
        corresponding correlation value) is calculated in specified time
        windows. Multiple time
        windows may be specified in ``tw``.

        :param ref_trc: Refernce trace for the shifting, defaults to None
        :type ref_trc: Optional[np.ndarray], optional
        :param tw: Time window(s) to check the shifting in, defaults to None.
        :type tw: Optional[List[float]], optional
        :param shift_range: Maximum shift range in seconds, defaults to 10
        :type shift_range: float, optional
        :param shift_steps: Number of shift steps, defaults to 101
        :type shift_steps: int, optional
        :type sides: str
        :param sides: Side of the traces to be used for the shifting estimate
            ('both' | 'single'). ``single`` is used for
            one-sided signals from active sources or if the time window shall
            not be symmetric. For ``both`` the time window will be mirrowd
            about
            zero lag time, e.g. [start,end] will result in time windows
            [-end:-start] and [start:end] being used simultaneousy
        :param return_sim_mat: Return simmilarity matrix?, defaults to False
        :type return_sim_mat: bool, optional
        :return: A DV object holding a shift value.
        :rtype: DV
        """
        if tw is not None:
            tw_list = [tw]
        dt = pcp.measure_shift(
            self.data, self.stats, ref_trc=ref_trc,
            tw=tw_list, shift_range=shift_range, shift_steps=shift_steps,
            sides=sides, return_sim_mat=return_sim_mat)[0]
        return dt

    def mirror(self):
        """
        Average the causal and acausal (i.e., right and left) parts of the
        correlation.

        ..note:: This action is performed **in-place**. If you would like to
            keep the original data use
            :func:`~seismic.correlate.stream.CorrelationBulk.copy()`.
        """
        self.data, self.stats = pcp.corr_mat_mirror(self.data, self.stats)
        self.stats.processing_bulk += ['Mirrored.']
        return self

    def resample(
        self, starttimes: List[UTCDateTime],
            endtimes: List[UTCDateTime] = []):
        """ Function to create correlation matrices with constant sampling

        When created from a CorrStream the correlation matrix contains all
        available correlation traces but homogeneous sampling is not guaranteed
        as correlation functions may be missing due to data gaps. This function
        restructures the correlation matrix by inserting or averaging
        correlation functions to provide temporally homogeneous sampling.
        Inserted correlation functions consist of 'nan' if gaps are present and
        averaging is done if more than one correlation function falls in a bin
        between start_times[i] and end_times[i]. If end_time is an empty list
        (default) end_times[i] is set to
        start_times[i] + (start_times[1] - start_times[0])

        :type start_times: list of class:`~obspy.core.UTCDateTime` objects
        :param start_times: list of starting times for the bins of the new
            sampling
        :type end_times: list of class:`~obspy.core.UTCDateTime` objects
        :param end_times: list of end times for the bins of the new
            sampling

        ..note:: This action is performed **in-place**. If you want to keep
            the original data use
            :func:`~seismic.correlate.stream.CorrelationBulk.copy()`
        """
        self.data, self.stats = pcp.corr_mat_resample(
            self.data, self.stats, starttimes, endtimes)
        self.stats.processing_bulk += [
            f'Resampled. Starttimes: {starttimes}, Endtimes: {endtimes}']
        return self

    def resample_time_axis(self, freq: float):
        """
        Resample the lapse time axis of a correlation matrix. The correlations
        are automatically filtered with a highpass filter of 0.4*sampling
        frequency to avoid aliasing. The function decides automatically whether
        to decimate or resample depending on the desired frequency

        :param freq: new sampling frequency
        :type freq: float

        ..note:: This action is performed **in-place**. If you want to keep
            the original data use
            :func:`~seismic.correlate.stream.CorrelationBulk.copy()`
        """
        self.data, self.stats = pcp.corr_mat_resample_or_decimate(
            self.data, self.stats, freq)
        self.stats.processing_bulk += [
            f'Resampled time axis. New sampling rate: {freq}Hz']
        return self

    def smooth(
        self, wsize: int, wtype: str = 'flat',
            axis: int = 1) -> np.ndarray:
        """
        Smoothes the correlation matrix with a given window function of the
        given width along the given axis. This method is based on the
        convolution of a scaled window with the signal. Each row/col
        (i.e. depending on the selected ``axis``) is "prepared" by introducing
        reflected copies of it (with the window size) in both ends so that
        transient parts are minimized in the beginning and end part of the
        resulting array.

        :type wsize: int
        :param wsize: Window size
        :type wtype: string
        :param wtype: Window type. It can be one of:
            ['flat', 'hanning', 'hamming', 'bartlett', 'blackman'] defaults to
            'flat'
        :type axis: int
        :param axis: Axis along with apply the filter. O: smooth along
            correlation lag time axis 1: smooth along time axis
        :rtype: :class:`~numpy.ndarray`
        :return: Filtered matrix

        ..note:: This action is performed **in-place**. If you want to keep
            the original data use
            :func:`~seismic.correlate.stream.CorrelationBulk.copy()`
        """
        self.data = pcp.corr_mat_smooth(self.data, wsize, wtype, axis)
        self.stats.processing_bulk += [
            f'Smoothed. wsize: {wsize}, wtype: {wtype}, axis: {axis}']
        return self

    def stretch(
        self, ref_trc: np.ndarray = None, tw: List[np.ndarray] = None,
        stretch_range: float = 0.1, stretch_steps: int = 101,
        sides: str = 'both', return_sim_mat: bool = False,
        ref_tr_trim: Optional[Tuple[float, float]] = None,
            ref_tr_stats=None, processing: Optional[dict] = None) -> DV:
        """
        Compute the velocity change with the stretching method
        (see Sens-Schönfelder and Wegler, 2006).

        :param ref_trc: Reference trace(s) to use for the computatio,
            defaults to None. Will extract a single trace if = None.
        :type ref_trc: np.ndarray, optional
        :param tw: Lapse Time window to use for the computation,
            defaults to None
        :type tw: List[np.ndarray], optional
        :param stretch_range: Maximum stretch value to test, defaults to 0.1
        :type stretch_range: float, optional
        :param stretch_steps: Number of stretch steps, defaults to 101
        :type stretch_steps: int, optional
        :param sides: Which sides to use. Can be 'left', 'right' (or 'single'),
            or 'both'. Defaults to 'both'
        :type sides: str, optional
        :param return_sim_mat: Return the similarity matrix, defaults to False
        :type return_sim_mat: bool, optional
        :param processing: dictionary holding processing information.
        :type processing: dict
        :return: The velocity change as :class:`~seismic.monitor.dv.DV` object.
        :rtype: DV
        """
        if ref_trc is None:
            ref_trc = self.ref_trc
        dv_dict = pcp.corr_mat_stretch(
            self.data, self.stats, ref_trc, tw, stretch_range, stretch_steps,
            sides, return_sim_mat, ref_tr_trim, ref_tr_stats)
        if not return_sim_mat:
            dv_dict['sim_mat'] = np.array([])
        return DV(**dv_dict, dv_processing=processing)

    def save(self, path: str):
        """
        Save the object to a numpy binary format (**.npz**)

        :param path: Output path
        :type path: str
        """
        kwargs = m3ut.save_header_to_np_array(self.stats)
        np.savez_compressed(
            path, data=self.data, **kwargs)

    def slice(
        self, starttime: UTCDateTime, endtime: UTCDateTime,
            include_partial: bool = True):
        """
        return a subset of the current CorrelationBulk with data inside
        the requested time window.

        :param starttime: Earliest Starttime
        :type starttime: UTCDateTime
        :param endtime: Latest endtime
        :type endtime: UTCDateTime
        :param include_partial: Include time window if it's only
            partially covered, defaults to True
        :type include_partial: bool, optional
        :return: Another CorrBulk object
        :rtype: CorrBulk
        """
        ii = self._find_slice_index(starttime, endtime, include_partial)
        data = self.data[ii, :]
        stats = deepcopy(self.stats)
        for key, value in stats.items():
            try:
                if isinstance(value, list) and key != 'processing_bulk':
                    stats[key] = [v for jj, v in zip(ii, value) if jj]
                elif isinstance(value, np.ndarray):
                    stats[key] = value[ii]
            except AttributeError:
                continue
                # Those will change automatically
        return CorrBulk(data, stats)

    def taper(self, width: float):
        """
        Taper the data.

        :param width: width to be tapered in seconds (per side)
        :type width: float

        ..note:: This action is performed **in-place**. If you want to keep
            the original data use
            :func:`~seismic.correlate.stream.CorrelationBulk.copy()`
        """
        self.data = pcp.corr_mat_taper(self.data, self.stats, width)
        proc = [f'tapered: width={width}s']
        self.stats.processing_bulk += proc
        return self

    def taper_center(self, width: float, slope_frac: float = 0.05):
        """
        Taper the central part of a correlation matrix.

        Due to electromagnetic cross-talk, signal processing or other effects
        the correlaton matrices are often contaminated around the zero lag
        time. This function tapers (multiples by zero) the central part of
        width `width`. To avoid problems with interpolation and filtering later
        on this is done with cosine taper.

        :param width: width of the central window to be tapered in seconds
            (**total length i.e., not per side**).
        :type width: float
        :param slope_frac: fraction of `width` used for soothing of edges,
            defaults to 0.05
        :type slope_frac: float, optional
        :return: self
        :rtype: CorrBulk

        ..note:: This action is performed **in-place**. If you want to keep
            the original data use
            :func:`~seismic.correlate.stream.CorrelationBulk.copy()`
        """
        self.data = pcp.corr_mat_taper_center(
            self.data, self.stats, width, slope_frac=slope_frac)
        proc = [f'tapered-centre: width={width}s, slope_frac={slope_frac}']
        self.stats.processing_bulk += proc
        return self

    def trim(self, starttime: float, endtime: float):
        """

        Trim the correlation matrix to the period from `starttime` to
        `endtime` given in seconds from the zero position, so both can be
        positive and negative.

        :type starttime: float
        :param starttime: start time in seconds with respect to the zero
            position
        :type endtime: float
        :param order: end time in seconds with respect to the zero position
        :return: self

        ..note:: This action is performed **in-place**. If you want to keep
            the original data use
            :func:`~seismic.correlate.stream.CorrelationBulk.copy()`
        """
        self.data, self.stats = corr_mat_trim(
            self.data, self.stats, starttime, endtime)
        proc = ['trim: %s, %s' % (str(starttime), str(endtime))]
        self.stats.processing_bulk += proc
        return self

    def wfc(
        self, ref_trc: np.ndarray, time_window: np.ndarray, sides: str,
        tw_start: float, tw_len: float, freq_min: float, freq_max: float,
            remove_nans: bool = True) -> WFC:
        """
        Computes the waveform coherency (**WFC**) between the given reference
        correlation(s) and correlation matrix.
        If several references are given, it will loop over these

        See Steinmann, et. al. (2021) for details.

        :param refcorr: 1 or 2D reference Correlation Trace extracted from the
            correlation data.
            If 2D, it will be interpreted as one refcorr trace per row.
        :type refcorr: np.ndarray
        :param tw: Lag time window to use for the computation
        :type tw: np.ndarray
        :param sides: Which sides to use. Can be `both`, `right`, `left`,
            or `single`.
        :type sides: str
        :param tw_start: Time window start in seconds lag time.
        :type tw_start: float
        :param tw_len: Length of the Lapse time window in seconds.
        :type tw_len: float
        :param freq_min: Highpass frequency used for the computed values. (Hz)
        :type freq_min: float
        :param freq_max: Lowpass frequency used for the computed values. (Hz)
        :type freq_max: float
        :param remove_nans: Remove nans from CorrMatrix, defaults to True
        :return: A dictionary with one correlation array for each reference
            trace. The keys are using the syntax `reftr_%n`.
        :rtype: dict
        """
        wfc_dict = wfc_multi_reftr(
            self.data, ref_trc, time_window, sides, remove_nans)
        wfc_processing = {
            'tw_start': tw_start,
            'tw_len': tw_len,
            'freq_min': freq_min,
            'freq_max': freq_max
        }
        wfc = WFC(wfc_dict, self.stats, wfc_processing)
        return wfc

    def _find_slice_index(
        self, starttime: UTCDateTime, endtime: UTCDateTime,
            include_partial: bool) -> np.ndarray:
        """
        Find the indices of the correlations in the requested time window.

        :param starttime: start of the time window
        :type starttime: UTCDateTime
        :param endtime: end of the time window
        :type endtime: UTCDateTime
        :param include_partial: include corrs that are only partially covered
        :type include_partial: bool
        :return: An Array containing the boolean indices
        :rtype: np.ndarray[bool]
        """
        if endtime <= starttime:
            raise ValueError('End has to be after start!')
        if include_partial:
            ii = np.logical_and(
                np.array(self.stats.corr_end) >= starttime,
                np.array(self.stats.corr_start) <= endtime)
        else:
            ii = np.logical_and(
                np.array(self.stats.corr_start) >= starttime,
                np.array(self.stats.corr_end) <= endtime)
        ii = np.squeeze(ii)
        if not len(np.nonzero(ii)[0]):
            warnings.warn(
                'No slices found in the requested time. Returning empty arr.')
        return ii


def read_corr_bulk(path: str) -> CorrBulk:
    """
    Reads a CorrBulk object from an **.npz** file.

    :param path: Path to file
    :type path: str
    :return: the corresponding and converted CorrBulk object
    :rtype: CorrBulk
    """
    loaded = np.load(path)
    stats = m3ut.load_header_from_np_array(loaded)
    return CorrBulk(loaded['data'], stats=stats)


class CorrStream(Stream):
    """
    Baseclass to hold correlation traces. Basically just a list of the
    correlation traces.
    """
    def __init__(self, traces: list = None):
        self.traces = []
        if isinstance(traces, CorrTrace):
            traces = [traces]
        if traces:
            for tr in traces:
                if not isinstance(tr, CorrTrace):
                    raise TypeError(
                        'Traces have to be of type'
                        ':class:`~seismic.correlate.correlate.CorrTrace`.')
                self.traces.append(tr)

    def __str__(self, extended=False) -> str:
        """
        Return short summary string of the current stream.

        It will contain the number of Traces in the Stream and the return value
        of each Trace's :meth:`~obspy.core.trace.Trace.__str__` method.

        :type extended: bool, optional
        :param extended: This method will show only 20 traces by default.
            Enable this option to show all entries.

        .. rubric:: Example

        >>> stream = Stream([Trace(), Trace()])
        >>> print(stream)  # doctest: +ELLIPSIS
        2 Trace(s) in Stream:
        ...
        """
        # get longest id
        if self.traces:
            id_length = self and max(len(tr.id) for tr in self) or 0
        else:
            id_length = 0
        out = str(len(self.traces)) + ' Correlation(s) in Stream:\n'
        if len(self.traces) <= 20 or extended is True:
            out = out + "\n".join([_i.__str__(id_length) for _i in self])
        else:
            out = out + "\n" + self.traces[0].__str__() + "\n" + \
                '...\n(%i other correlations)\n...\n' % (len(self.traces) - 2)\
                + self.traces[-1].__str__() + '\n\n[Use "print(' + \
                'Stream.__str__(extended=True))" to print all correlaitons]'
        return out

    def create_corr_bulk(
        self, network: str = None, station: str = None, channel: str = None,
        location: str = None, times: Tuple[UTCDateTime, UTCDateTime] = None,
            inplace=True) -> CorrBulk:
        """
        Creates a CorrelationBulk object, which offers additional options for
        faster postprocessing.

        :param network: Select only this network, defaults to None
        :type network: str, optional
        :param station: Take data from this station, defaults to None
        :type station: str, optional
        :param channel: Take data from this channel. If None all channels will
            be retained, defaults to None
        :type channel: str, optional
        :param location: Take data from only this location. Else various
            locations can be processed together, defaults to None
        :type location: str, optional
        :param times: Only take data from this time window, defaults to None
        :type times: Tuple[UTCDateTime, UTCDateTime], optional
        :param inplace: The original data will be deleted to save memory,
            defaults to True.
        :type inplace: bool, optional
        :return: The CorrelationBulk object
        :rtype: CorrBulk

        .. note:: This function will check whether the metadata of the input
            stream is identical, so that correlations from different stations,
            components, or differently processed data cannot be mixed.
        """
        st = self.select(network, station, location, channel)
        if times:
            st = st.select_corr_time(times[0], times[1])
        A = np.empty((st.count(), st[0].stats.npts))
        statlist = []
        # Double check sampling rate
        sr = st[0].stats.sampling_rate
        mask = []
        for ii, tr in enumerate(st):
            if tr.stats.sampling_rate != sr:
                warnings.warn('Sampling rate differs. Trace is skipped.')
                mask.append(ii)
                continue
            A[ii] = tr.data
            if inplace:
                del tr.data
            statlist.append(tr.stats)
        # Apply mask for skipped data
        A = np.delete(A, mask, 0)

        if channel is None or '*' in channel or '?' in channel:
            stats = convert_statlist_to_bulk_stats(
                statlist, varying_channel=True)
        else:
            stats = convert_statlist_to_bulk_stats(statlist)
        return CorrBulk(A, stats)

    def plot(
        self, sort_by: str = 'corr_start',
        timelimits: Tuple[float, float] = None,
        ylimits: Tuple[float, float] = None, scalingfactor: float = None,
        ax: plt.Axes = None, linewidth: float = 0.25,
        outputfile: str = None, title: str = None, type: str = 'heatmap',
            cmap: str = 'inferno', vmin: float = None, vmax: float = None,
            **kwargs):
        """
        Creates a section plot of all correlations in this stream.
        kwargs will be passed to
        :func:`~seismic.plot.plot_correlation.plot_cst`

        :param sort_by: Which parameter to plot against. Can be either
            ``corr_start`` or ``distance``, defaults to 'corr_start'.
        :type sort_by: str, optional
        :param timelimits: xlimits (lag time) in seconds, defaults to None
        :type timelimits: Tuple[float, float], optional
        :param ylimits: limits for Y-axis (either a :class:`datetime.datetime`
            or float in km (if plotted against distance)), defaults to None.
        :type ylimits: Tuple[float, float], optional
        :param scalingfactor: Which factor to scale the Correlations with. Play
            around with this if you want to make amplitudes bigger or smaller,
            defaults to None (automatically chosen).
        :type scalingfactor: float, optional
        :param ax: Plot in existing axes? Defaults to None
        :type ax: plt.Axes, optional
        :param linewidth: Width of the lines to plot, defaults to 0.25
        :type linewidth: float, optional
        :param outputfile: Save the plot? defaults to None
        :type outputfile: str, optional
        :param title: Title of the plot, defaults to None
        :type title: str, optional
        :param type: Type of plot. Either `'heatmap'` for a heat plot or
            `'section'` for a wiggle type plot. Defaults to heatmap.
        :type title: str, optional
        :param cmap: Decides about colormap if type == 'heatmap'.
        Defaults to 'inferno'.
        :type cmap: str, optional

        .. note:: If you would like to plot a subset of this stream, use
            :func:`~seismic.correlate.stream.CorrStream.select`.
        """
        if self.count() == 1:
            return self[0].plot(tlim=timelimits)
        ax = plot_cst(
            self, sort_by=sort_by, timelimits=timelimits, ylimits=ylimits,
            scalingfactor=scalingfactor, ax=ax, linewidth=linewidth,
            outputfile=outputfile, title=title, type=type, cmap=cmap,
            vmin=vmin, vmax=vmax, **kwargs)
        return ax

    def pop_at_utcs(self, utcs: npt.NDArray[UTCDateTime]):
        """
        Remove Correlations that contain any time given in ``utcs``.

        :param utcs: Array Containing UTC times that should be filtered.
            If UTC is between any of corr_start and corr_end, the corresponding
            CorrTrace will be removed from the stream.
        :type utcs: npt.ArrayLike[UTCDateTime]
        :return: filtered CorrStream
        :rtype: CorrStream
        """
        cst_filt = CorrStream()
        for ctr in self:
            if not np.any(np.all(
                [utcs > ctr.stats.corr_start, utcs < ctr.stats.corr_end],
                    axis=0)):
                cst_filt.append(ctr)
        return cst_filt

    def remove_duplicates(self):
        """
        Removes identical CorrTraces.

        .. note:: This action is performed **in-place**. If you want to keep
            the original data use
            :func:`~seismic.correlate.stream.CorrStream.copy()`
        """
        cst_filt = CorrStream()
        for ctr in self:
            if ctr not in cst_filt:
                cst_filt.append(ctr)
        self.traces = cst_filt.traces

    def select_corr_time(
        self, starttime: UTCDateTime, endtime: UTCDateTime,
            include_partially_selected: bool = True):
        """
        Selects correlations that are inside of the requested time window.

        :param starttime: Requested start
        :type starttime: UTCDateTime
        :param endtime: Requested end
        :type endtime: UTCDateTime
        :param include_partially_selected: If set to ``True``, also the half
            selected time window **before** the requested time will be attached
            Given the following stream containing 6 correlations, "|" are the
            correlation starts and ends, "A" is the requested starttime and "B"
            the corresponding endtime::

                |         |A        |         |       B |         |
                1         2         3         4         5         6

            ``include_partially_selected=True`` will select samples 2-4,
            ``include_partially_selected=False`` will select samples 3-4 only.
            Defaults to True
        :type include_partially_selected: bool, optional
        :return: Correlation Stream holding all selected traces
        :rtype: CorrStream
        """
        self.sort(keys=['corr_start'])
        outst = CorrStream()
        # the 2 seconds difference are to avoid accidental smoothing
        if include_partially_selected:
            for tr in self:
                if (tr.stats.corr_end > starttime
                    and tr.stats.corr_end < endtime) \
                        or tr.stats.corr_end == endtime:
                    outst.append(tr)
            return outst
        # else
        for tr in self:
            if tr.stats.corr_start >= starttime \
                    and tr.stats.corr_end <= endtime:
                outst.append(tr)
        return outst

    def slide(
        self, window_length: float, step: float,
        include_partially_selected: bool = True,
            starttime: UTCDateTime = None, endtime: UTCDateTime = None):
        """
        Generator yielding correlations that are inside of each requested time
        window and inside of this stream.

        Please keep in mind that it only returns a new view of the original
        data. Any modifications are applied to the original data as well. If
        you don't want this you have to create a copy of the yielded
        windows. Also be aware that if you modify the original data and you
        have overlapping windows, all following windows are affected as well.

        Not all yielded windows must have the same number of traces. The
        algorithm will determine the maximal temporal extents by analysing
        all Traces and then creates windows based on these times.


        :param window_length: The length of the requested time window in
            seconds. Note that the window length has to correspond at least to
            the length of the longest correlation window (i.e., the length of
            the correlated waveforms). This is because the correlations cannot
            be sliced.
        :type window_length: float
        :param step: The step between the start times of two successive
            windows in seconds. Has to be greater than 0
        :type step: float
        :param include_partially_selected: If set to ``True``, also the half
            selected time window **before** the requested time will be attached
            Given the following stream containing 6 correlations, "|" are the
            correlation starts and ends, "A" is the requested starttime and "B"
            the corresponding endtime::

                |         |A        |         |       B |         |
                1         2         3         4         5         6

            ``include_partially_selected=True`` will select samples 2-4,
            ``include_partially_selected=False`` will select samples 3-4 only.
            Defaults to True.
        :type include_partially_selected: bool, optional
        :param starttime: Start the sequence at this time instead of the
            earliest available starttime.
        :type starttime: UTCDateTime
        :param endtime: Start the sequence at this time instead of the
            latest available endtime.
        :type endtime: UTCDateTime
        """
        if starttime is None:
            starttime = min(tr.stats.corr_start for tr in self)
        if endtime is None:
            endtime = max(tr.stats.corr_end for tr in self)

        if window_length < max(
                tr.stats.corr_end-tr.stats.corr_start for tr in self):
            raise ValueError(
                'The length of the requested time window has to be larger or'
                'equal than the actual correlation length of one window.'
                'i.e., correlations can not be sliced, only selected.')

        if step <= 0:
            raise ValueError('Step has to be larger than 0.')

        windows = np.arange(
            starttime.timestamp, endtime.timestamp, step)

        if len(windows) < 1:
            return

        for start in windows:
            start = UTCDateTime(start)
            stop = start + window_length
            temp = self.select_corr_time(
                start, stop,
                include_partially_selected=include_partially_selected)
            # It might happen that there is a time frame where there are no
            # windows, e.g. two traces separated by a large gap.
            if not temp:
                continue
            yield temp

    def stack(
        self, weight: str = 'by_length', starttime: UTCDateTime = None,
        endtime: UTCDateTime = None, stack_len: int | str = 0,
            regard_location=True):
        """
        Average the data of all traces in the given time windows.
        Will only stack data from the same network/channel/station combination.
        Location codes will only optionally be regarded.

        :param starttime: starttime of the stacking time windows. If None, the
            earliest available is chosen, defaults to None.
        :type starttime: UTCDateTime, optional
        :param endtime: endtime of the stacking time windows. If None, the
            latest available is chosen, defaults to None
        :type endtime: UTCDateTime, optional
        :param stack_len: Length of one stack. Is either a value in seconds,
            the special option "daily" (creates 24h stacks that always start at
            midnight), or 0 for a single stack over the whole time period,
            defaults to 0.
        :type stack_len: intorstr, optional
        :param regard_location: Don't stack correlations with varying location
            code combinations, defaults to True.
        :type regard_location: bool, optional
        :return: A stream holding the stacks.
        :rtype: :class`~seismic.correlate.stream.CorrStream`
        """

        # Seperate if there are different stations channel and or locations
        # involved
        if stack_len == 0:
            return stack_st_by_group(self, regard_location, weight)

        # else
        self.sort(keys=['corr_start'])
        if not starttime:
            starttime = self[0].stats.corr_start
        if not endtime:
            endtime = self[-1].stats.corr_end
        outst = CorrStream()
        if stack_len == 'daily':
            starttime = UTCDateTime(
                year=starttime.year, julday=starttime.julday)
            stack_len = 3600*24
        for st in self.slide(
                stack_len, stack_len, include_partially_selected=True,
                starttime=starttime, endtime=endtime):
            outst.extend(stack_st_by_group(st, regard_location, weight))
        return outst

    def _to_matrix(
        self, network: str = None, station: str = None, channel: str = None,
        location: str = None,
        times: Tuple[UTCDateTime, UTCDateTime] = None) -> Tuple[
            np.ndarray, Iterator[Stats]]:
        """
        Creates a numpy array from the data in the
        :class:`~seismic.correlate.stream.Stream` object. Also returns a list
        of the Stats objects. The positional arguments are filter arguments.

        :param st: Input Stream
        :type st: CorrStream
        :return: both a numpy array (i.e., matrix) and a list of the stats
        :rtype: Tuple[np.ndarray, Iterator[Stats]]
        """
        st = self.select(network, station, location, channel)
        if times:
            st = st.select_corr_time(times[0], times[1])

        st.sort(keys=['corr_start'])
        A = np.empty((st.count(), st[0].stats.npts))
        stats = []
        for ii, tr in enumerate(st):
            A[ii] = tr.data
            stats.append(tr.stats)
        return A, stats


class CorrTrace(Trace):
    """
    Baseclass to hold correlation data. Derived from the class
    :class:`~obspy.core.trace.Trace`.
    """
    def __init__(
        self, data: np.ndarray, header1: Stats = None,
        header2: Stats = None, inv: Inventory = None,
        start_lag: float = None, end_lag: float = None,
            _header: dict = None):
        """
        Initialise the correlation trace. Is done by combining the stats of the
        two :class:`~obspy.core.trace.Trace` objects' headers. If said headers
        do not contain Station information (i.e., coordinates), an
        :class:`~obspy.core.inventory.Inventory` with information about both
        stations should be provided as well.

        :param data: The correlation data
        :type data: np.ndarray
        :param header1: header of the first trace, defaults to None
        :type header1: Stats, optional
        :param header2: header of the second trace, defaults to None
        :type header2: Stats, optional
        :param inv: Inventory object for the stations, defaults to None
        :type inv: Inventory, optional
        :param start_lag: The lag of the first sample of the correlation given
            in seconds.
        :type start_lag: float
        :param end_lag: The lag of the last sample of the correlation
            in seconds.
        :type end_lag: float
        :param _header: Already combined header, used when reading correlations
            from a file, defaults to None
        :type _header: dict, optional
        """
        if _header:
            header = CorrStats(_header)
        elif not header1 and not header2:
            header = CorrStats()
            if start_lag:
                header['start_lag'] = start_lag
        else:
            header, data = alphabetical_correlation(
                header1, header2, start_lag, end_lag, data, inv)

        super(CorrTrace, self).__init__(data=data)
        self.stats = header
        self.stats['npts'] = len(data)

    def __str__(self, id_length: int = None) -> str:
        """
        Return short summary string of the current trace.

        :rtype: str
        :return: Short summary string of the current trace containing the SEED
            identifier, start time, end time, sampling rate and number of
            points of the current trace.

        .. rubric:: Example

        >>> tr = Trace(header={'station':'FUR', 'network':'GR'})
        >>> str(tr)  # doctest: +ELLIPSIS
        'GR.FUR.. | 1970-01-01T00:00:00.000000Z - ... | 1.0 Hz, 0 samples'
        """
        # set fixed id width
        if id_length:
            out = "%%-%ds" % (id_length)
            trace_id = out % self.id
        else:
            trace_id = "%s" % self.id
        out = ''
        # output depending on delta or sampling rate bigger than one
        if self.stats.sampling_rate < 0.1:
            if hasattr(self.stats, 'preview') and self.stats.preview:
                out = out + ' | '\
                    "%(corr_start)s - %(corr_end)s | " + \
                    "%(delta).1f s, %(npts)d samples [preview]"
            else:
                out = out + ' | '\
                    "%(corr_start)s - %(corr_end)s | " + \
                    "%(delta).1f s, %(npts)d samples"
        else:
            if hasattr(self.stats, 'preview') and self.stats.preview:
                out = out + ' | '\
                    "%(corr_start)s - %(corr_end)s | " + \
                    "%(sampling_rate).1f Hz, %(npts)d samples [preview]"
            else:
                out = out + ' | '\
                    "%(corr_start)s - %(corr_end)s | " + \
                    "%(sampling_rate).1f Hz, %(npts)d samples"
        # check for masked array
        if np.ma.count_masked(self.data):
            out += ' (masked)'
        return trace_id + out % (self.stats)

    def plot(
        self, tlim: Tuple[float, float] = None, ax: plt.Axes = None,
            outputdir: str = None, clean: bool = False) -> plt.Axes:
        """
        Plots thios CorrelationTrace.

        :param tlim: Limits for the lapse axis in seconds, defaults to None
        :type tlim: Tuple[float, float], optional
        :param ax: Plot in existing axes, defaults to None
        :type ax: plt.Axes, optional
        :param outputdir: Save this plot? Defaults to None
        :type outputdir: str, optional
        :param clean: Make a clean plot without labels & axes,
            defaults to False.
        :type clean: bool, optional
        """
        return plot_ctr(self, tlim, ax, outputdir, clean)

    def times(self) -> np.ndarray:
        """
        Convenience Function that returns an array holding the lag times of the
        correlation.

        :return: Array with lag times
        :rtype: np.ndarray
        """
        # this way there won't be problems with numerical errors for small
        # delta
        return np.arange(self.stats.npts)*self.stats.delta\
            + self.stats.start_lag


def alphabetical_correlation(
    header1: Stats, header2: Stats, start_lag: float, end_lag: float,
        data: np.ndarray, inv: Inventory) -> Tuple[CorrStats, np.ndarray]:
    """
    Make sure that Correlations are always created in alphabetical order,
    so that we won't have both a correlation for AB-CD and CD-AB.
    If the correlation was computed in the wrong order, the corr-data will be
    flipped along the t-axis.

    :param header1: Header of the first trace.
    :type header1: Stats
    :param header2: Header of the second trace
    :type header2: Stats
    :param start_lag: start lag in s
    :type start_lag: float
    :param end_lag: end lag in s
    :type end_lag: float
    :param data: The computed cross-correlation for header1-header2
    :type data: np.ndarray
    :param inv: The inventory holding the station coordinates. Only needed if
        coords aren't provided in stats.
    :type inv: Inventory
    :return: the header for the CorrTrace and the data
        (will also be modified in place)
    :rtype: Tuple[Stats, np.ndarray]
    """
    # make sure the order is correct
    # Will do that always alphabetically sorted
    sort1 = header1.network + header1.station + header1.channel
    sort2 = header2.network + header2.station + header2.channel
    sort = [sort1, sort2]
    sorted = sort.copy()
    sorted.sort()
    if sort != sorted:
        header = combine_stats(
            header2, header1, -end_lag,
            inv=inv)
        # reverse array and lag times
        data = np.flip(data)
    else:
        header = combine_stats(
            header1, header2, start_lag,
            inv=inv)
    return header, data


def combine_stats(
    stats1: Stats, stats2: Stats, start_lag: float,
        inv: Inventory = None) -> CorrStats:
    """ Combine the meta-information of two ObsPy Trace.Stats objects

    This function returns a ObsPy :class:`~obspy.core.trace.Stats` object
    obtained combining the two associated with the input Traces.
    Namely ``stats1`` and ``stats2``.

    The fields ['network','station','location','channel'] are combined in
    a ``-`` separated fashion to create a "pseudo" SEED like ``id``.

    For all the others fields, only "common" information are retained: This
    means that only keywords that exist in both dictionaries will be included
    in the resulting one.

    :type stats1: :class:`~obspy.core.trace.Stats`
    :param stats1: First Trace's stats
    :type stats2: :class:`~obspy.core.trace.Stats`
    :param stats2: Second Trace's stats
    :param start_lag: The lag of the first sample of the correlation given
        in seconds (usually negative).
    :type start_lag: float
    :type inv: :class:`~obspy.core.inventory.Inventory`, optional
    :param inv: Inventory containing the station coordinates. Only needed if
        station coordinates are not in Trace.Stats. Defaults to None.

    :rtype: :class:`~obspy.core.trace.Stats`
    :return: **stats**: combined Stats object
    """

    if not isinstance(stats1, Stats):
        raise TypeError("stats1 must be an obspy Stats object.")

    if not isinstance(stats2, Stats):
        raise TypeError("stats2 must be an obspy Stats object.")

    # We also have to remove these as they are obspy AttributeDicts as well
    stats1.pop('asdf', None)
    stats2.pop('asdf', None)

    tr1_keys = list(stats1.keys())
    tr2_keys = list(stats2.keys())

    stats = CorrStats()
    # actual correlation times
    stats['corr_start'] = max(stats1.starttime, stats2.starttime)
    stats['corr_end'] = min(stats1.endtime, stats2.endtime)

    # Adjust the information to create a new SEED like id
    keywords = ['network', 'station', 'location', 'channel']
    sac_keywords = ['sac']

    for key in keywords:
        if key in tr1_keys and key in tr2_keys:
            stats[key] = stats1[key] + '-' + stats2[key]

    for key in tr1_keys:
        if key not in keywords and key not in sac_keywords:
            if key in tr2_keys:
                if stats1[key] == stats2[key]:
                    # in the stats object there are read only objects
                    try:
                        stats[key] = stats1[key]
                    except (AttributeError, KeyError):
                        pass

    try:
        if ('stla' and 'stlo' and 'stel') in stats1:
            stats['stla'] = stats1.stla
            stats['stlo'] = stats1.stlo
            stats['stel'] = stats1.stel
            stats['evla'] = stats2.stla
            stats['evlo'] = stats2.stlo
            stats['evel'] = stats2.stel
        else:
            stats['stla'] = stats1.sac.stla
            stats['stlo'] = stats1.sac.stlo
            stats['stel'] = stats1.sac.stel
            stats['evla'] = stats2.sac.stla
            stats['evlo'] = stats2.sac.stlo
            stats['evel'] = stats2.sac.stel
            stats1.update(stats1['sac'])
            stats2.update(stats2['sac'])

        az, baz, dist = m3ut.trace_calc_az_baz_dist(stats1, stats2)

        stats['dist'] = dist / 1000
        stats['az'] = az
        stats['baz'] = baz
    except (AttributeError, KeyError):
        if inv:
            inv1 = inv.select(
                network=stats1.network, station=stats1.station)
            inv2 = inv.select(
                network=stats2.network, station=stats2.station)
            stats['stla'] = inv1[0][0].latitude
            stats['stlo'] = inv1[0][0].longitude
            stats['stel'] = inv1[0][0].elevation
            stats['evla'] = inv2[0][0].latitude
            stats['evlo'] = inv2[0][0].longitude
            stats['evel'] = inv2[0][0].elevation

            az, baz, dist = m3ut.inv_calc_az_baz_dist(inv1, inv2)

            stats['dist'] = dist / 1000
            stats['az'] = az
            stats['baz'] = baz
        else:
            warnings.warn("No station coordinates provided.")
    stats.pop('sac', None)
    stats.pop('response', None)
    stats['_format'] = 'hdf5'

    # note that those have to be adapted whenever several correlations are
    # stacked
    stats['start_lag'] = start_lag
    return stats


Compare_Str = "{network}.{station}.{channel}.{location}"
Compare_Str_No_Loc = "{network}.{station}.{channel}"


def compare_tr_id(tr0: Trace, tr1: Trace, regard_loc: bool = True) -> bool:
    """
    Check whether two traces are from the same channel, station, network, and,
    optionally, location. Useful for stacking

    :param tr0: first trace
    :type tr0: :class:`~obspy.core.trace.Trace`
    :param tr1: second trace
    :type tr1: :class:`~obspy.core.trace.Trace`
    :param regard_loc: Regard the location code or not
    :type regard_loc: bool
    :return: Bool whether the two are from the same (True) or not (False)
    :rtype: bool
    """
    if regard_loc:
        return Compare_Str.format(**tr0.stats) \
            == Compare_Str.format(**tr1.stats)
    else:
        return Compare_Str_No_Loc.format(**tr0.stats) \
            == Compare_Str_No_Loc.format(**tr1.stats)


def stack_st_by_group(st: Stream, regard_loc: bool, weight: str) -> CorrStream:
    """
    Stack all traces that belong to the same network, station, channel, and
    (optionally) location combination in the input stream.

    :param st: input Stream
    :type st: Stream
    :param regard_loc: Seperate data with different location code
    :type regard_loc: bool
    :return: :class:`~seismic.correlate.stream.CorrStream`
    :rtype: CorrStream
    """
    if regard_loc:
        key = "{network}.{station}.{channel}.{location}"
    else:
        key = "{network}.{station}.{channel}"
    stackdict = {}
    for tr in st:
        stackdict.setdefault(key.format(**tr.stats), CorrStream()).append(tr)
    stackst = CorrStream()
    for k in stackdict:
        stackst.append(stack_st(stackdict[k], weight))
    return stackst


def stack_st(st: CorrStream, weight: str, norm: bool = True) -> CorrTrace:
    """
    Returns an average of the data of all traces in the stream. Also adjusts
    the corr_start and corr_end parameters in the header.

    :param st: input Stream
    :type st: CorrStream
    :param weight: type of weigthing to use. Either `mean` or `by_length`
    :type weigth: str
    :param norm: Should the traces be normalised by their absolute maximum
        prior to stacking?
    :type norm: bool
    :return: Single trace with stacked data
    :rtype: CorrTrace
    """
    st.sort(keys=['corr_start'])
    stats = st[0].stats.copy()
    stats['corr_end'] = st[-1].stats['corr_end']
    st.sort(keys=['npts'])
    npts = st[-1].stats.npts
    stack = []
    dur = []  # duration of each trace
    for tr in st.select(npts=npts):
        stack.append(tr.data)
        dur.append(tr.stats.corr_end-tr.stats.corr_start)
    A = np.array(stack)
    if weight == 'mean' or weight == 'average':
        data = np.nanmean(A, axis=0)
        if norm:
            norm = np.nanmax(np.abs(data))
            data /= norm  # np.tile(np.atleast_2d(norm).T, (1, A.shape[1]))
        return CorrTrace(data, _header=stats)
    elif weight == 'by_length':
        # Weight by the length of each trace
        data = np.nansum((A.T*np.array(dur)).T, axis=0)/np.nansum(dur)
        if norm:
            norm = np.nanmax(np.abs(data))
            data /= norm  # np.tile(np.atleast_2d(norm).T, (1, A.shape[1]))
        return CorrTrace(data=data, _header=stats)


def convert_statlist_to_bulk_stats(
        statlist: List[CorrStats], varying_loc: bool = True,
        varying_channel: bool = False) -> CorrStats:
    """
    Converts a list of :class:`~seismic.correlate.stream.CorrTrace` stats
    objects to a single stats object that can be used for the creation of a
    :class:`~seismic.correlate.stream.CorrBulk` object

    :param statlist: list of Stats
    :type statlist: List[Stats]
    :param varying_loc: Set true if the location codes vary. Defaults to True.
    :type varying_loc: bool
    :param varying_loc: Set true if the channel codes vary (e.g., EHZ to BHZ),
        defaults to False.
    :type varying_loc: bool
    :raises ValueError: raised if data does not fit together
    :return: single Stats object
    :rtype: Stats
    """
    stats = statlist[0].copy()
    # can change from trace to trace
    mutables = ['corr_start', 'corr_end']

    # Should / have to be identical for each trace
    # Not 100% sure if start and end_lag should be on this list
    immutables = [
        'npts', 'sampling_rate', 'network', 'station', 'start_lag',
        'end_lag', 'stla', 'stlo', 'stel', 'evla', 'evlo', 'evel',
        'dist', 'az', 'baz']
    if varying_loc:
        mutables += ['location']
    else:
        immutables += ['location']
    if varying_channel:
        mutables += ['channel']
    else:
        immutables += ['channel']
    for key in mutables:
        stats[key] = []
    for trstat in statlist:
        for key in mutables:
            stats[key] += [trstat[key]]
        for key in immutables:
            try:
                if stats[key] != trstat[key]:
                    raise ValueError(
                        'The stream contains data with different properties. '
                        + f'The differing property is {key}. '
                        + f'With the values: {stats[key]} and {trstat[key]}. '
                        + f'For station {stats["network"]}.{stats["station"]}')
            except KeyError:
                warnings.warn(f'No information about {key} in header.')
    stats['ntrcs'] = len(statlist)
    return stats



'''
:copyright:
   The SeisMIC development team (makus@gfz-potsdam.de).
:license:
    EUROPEAN UNION PUBLIC LICENCE v. 1.2
   (https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12)
:author:
   Peter Makus (makus@gfz-potsdam.de)

Created: Friday, 5th November 2021 08:19:58 am
Last Modified: Tuesday, 11th July 2023 04:42:52 pm
'''
import glob
from typing import List

import numpy as np
from matplotlib import pyplot as plt

from seismic.correlate.stats import CorrStats
from seismic.utils import miic_utils as mu
from seismic.plot.plot_wfc import plot_wfc_bulk


class WFC(dict):
    """
    Object to hold the properties of a waveform coherency for a single
    frequency and lapse time window.

    The waveform coherency can be used as a measure of stability of
    a certain correlation. See Steinmann, et. al. (2021) for details.
    """
    def __init__(
            self, wfc_dict: dict, stats: CorrStats, wfc_processing: dict):
        """
        Initialise WFC object.

        :param wfc_dict: Dictionary holding the correlation values for
            each of the reference traces. The keys follow the syntax
            `reftr_%n` where n is the number of the reference trace
            (starting with 0).
        :type wfc_dict: dict
        :param stats: CorrStats object from the original CorrBulk.
            Also has to contain the keys holding the information about the
            bandpass filter and the used lapse time window.
        :type stats: CorrStats
        """
        for k, v in wfc_dict.items():
            self[k] = v
        del wfc_dict
        # dict to hold averages
        self.av = {}
        self.stats = stats
        self.mean = np.array([])
        self.wfc_processing = wfc_processing

    def compute_average(self):
        """
        Computes the average of the waveform coherency over the time axis.

        :param average_reftrcs: Average over all reference traces. Defaults to
            True.
        :type average_reftrcs: bool, optional
        """
        for k, v in self.items():
            self.av['%s_av' % k] = np.nanmean(v)
        self.mean = np.nanmean([v for v in self.av.values()])

    def save(self, path: str):
        """
        Saves the object to an npz file.

        :param path: Path to save to.
        :type path: str
        """
        kwargs = mu.save_header_to_np_array(self.stats)
        kwargs.update(self.wfc_processing)
        np.savez_compressed(path, mean=self.mean, **self.av, **self, **kwargs)


class WFCBulk(object):
    """
    Object to hold the waveform coherency for one station and subsequently plot
    it against frequency and lapse time window.
    """
    def __init__(self, wfcl: List[WFC]):
        # Compute averages if not available
        freq = []
        lw = []
        means = []
        for wfc in wfcl:
            wfc.mean.size or wfc.compute_average()
            # Midpoint of bp-filter (centre frequency)
            freq.append(
                (wfc.wfc_processing['freq_max']
                 + wfc.wfc_processing['freq_min']) / 2)
            # midpoint of lapse window
            lw.append(
                wfc.wfc_processing['tw_start']
                + wfc.wfc_processing['tw_len']/2)
            means.append(wfc.mean)
        # create grids
        lwg, freqg = np.meshgrid(sorted(set(lw)), sorted(set(freq)))
        # actual wfc grid
        self.wfc = np.full_like(lwg, np.nan)
        for f, lwc, v in zip(freq, lw, means):
            ii = np.argwhere(((lwg == lwc) & (freqg == f)))[0]
            self.wfc[ii[0], ii[1]] = v
        del means
        # create grids
        self.lw = np.array(sorted(set(lw)))
        self.lwg = lwg
        del lw
        self.cfreq = np.array(sorted(set(freq)))
        self.cfreqg = freqg
        del freq

    def plot(
        self, title: str = None, log: bool = False,
            cmap: str = 'viridis') -> plt.Axes:
        """
        Create a plot of the waveform coherency against frequency and lapse
        time window.

        :param title: Title of the plot, defaults to None
        :type title: str, optional
        :param log: plot on logarithmic frequency axis, defaults to False
        :type log: bool, optional
        """
        return plot_wfc_bulk(
            self.lw, self.cfreq, self.wfc, title=title,
            log=log, cmap=cmap)


def read_wfc(path: str) -> WFC:
    """
    Read a :class:`~seismic.monitor.wfc.WFC` object saved as *npz*.
    Returns a list if wildcards are used

    :param path: Path to file
    :type path: str
    :return: The WFC object.
    :rtype: WFC
    """
    if '*' in path:
        return [read_wfc(f) for f in glob.glob(path, recursive=True)]
    loaded = np.load(path)
    d = {}
    av = {}
    stats = {}
    for k, v in loaded.items():
        if k[-3:] == '_av':
            av[k] = v
        elif k == 'mean':
            mean = v
        elif 'reftr' in k:
            d[k] = v
        else:
            stats[k] = v
    wfc_processing = {
        'tw_start': stats['tw_start'],
        'tw_len': stats['tw_len'],
        'freq_min': stats['freq_min'],
        'freq_max': stats['freq_max']
    }
    wfc = WFC(
        d, stats=CorrStats(mu.load_header_from_np_array(stats)),
        wfc_processing=wfc_processing)
    wfc.av = av
    wfc.mean = mean
    return wfc



'''
Module containing functions for preprocessing in the frequency domain


:copyright:
   The SeisMIC development team (makus@gfz-potsdam.de).
:license:
    EUROPEAN UNION PUBLIC LICENCE v. 1.2
   (https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12)
:author:
   Peter Makus (makus@gfz-potsdam.de)

Created: Tuesday, 20th July 2021 03:40:11 pm
Last Modified: Wednesday, 28th June 2023 02:00:00 pm
'''
from copy import deepcopy
import logging

import numpy as np
import obspy.signal as osignal


def FDfilter(B: np.ndarray, args: dict, params: dict) -> np.ndarray:
    """
    Filter Fourier-transformed data

    Filter Fourier tranformed data by tapering in frequency domain. The `args`
    dictionary is supposed to contain the key `flimit` with a value that is a
    four element list or tuple defines the corner frequencies (f1, f2, f3, f4)
    in Hz of the cosine taper which is one between f2 and f3 and tapers to zero
    for f1 < f < f2 and f3 < f < f4.

    :type B: numpy.ndarray
    :param B: Fourier transformed time series data with frequency oriented\\
        along the first dimension (columns)
    :type args: dictionary
    :param args: arguments dictionary as described above
    :type params: dictionary
    :param params: params['freqs'] contains an array with the freqency values
        of the samples in `B`

    :rtype: numpy.ndarray
    :return: filtered spectal data
    """
    args = deepcopy(args)
    args.update({'freqs': params['freqs']})
    tap = osignal.invsim.cosine_taper(B.shape[1], **args)
    B *= np.tile(np.atleast_2d(tap), (B.shape[0], 1))
    return B


def FDsignBitNormalization(
        B: np.ndarray, args: dict, params: dict) -> np.ndarray:
    """
    Sign bit normalization of frequency transformed data

    Divides each sample by its amplitude resulting in trace with amplidues of
    (-1, 0, 1). As this operation is done in frequency domain it requires two
    Fourier transforms and is thus quite costly but alows to be performed
    after other steps of frequency domain procesing e.g. whitening.


    :type B: numpy.ndarray
    :param B: Fourier transformed time series data with frequency oriented\\
        along the first dimension (columns)
    :type args: dictionary
    :param args: not used in this function
    :type params: dictionary
    :param params: not used in this function

    :rtype: numpy.ndarray
    :return: frequency transform of the 1-bit normalized data
    """
    B = np.fft.irfft(B)
    # np.sign only takes the real part into account
    C = np.sign(B)
    return np.fft.rfft(C)


def spectralWhitening(B: np.ndarray, args: dict, params) -> np.ndarray:
    """
    Spectal whitening of Fourier-transformed date

    Normalize the amplitude spectrum of the complex spectra in `B`. The
    `args` dictionary may contain the keyword `joint_norm`. If its value is
    True the normalization of sets of three traces are normalized jointly by
    the mean of their amplitude spectra. This is useful for later rotation of
    correlated traces in the ZNE system into the ZRT system.

    :type B: numpy.ndarray
    :param B: Fourier transformed time series data with frequency oriented\\
        along the first dimension (columns)
    :type args: dictionary
    :param args: arguments dictionary as described above
    :type params: dictionary
    :param params: not used here

    :rtype: numpy.ndarray
    :return: whitened spectal data
    """
    absB = np.absolute(B)
    if 'joint_norm' in list(args.keys()):
        if args['joint_norm']:
            assert B.shape[0] % 3 == 0, "for joint normalization the number\
                      of traces needs to the multiple of 3: %d" % B.shape[1]
            for ii in np.arange(0, B.shape[0], 3):
                absB[ii:ii+3, :] = np.tile(
                    np.atleast_2d(np.mean(absB[ii:ii+3, :], axis=0)), [3, 1])
    with np.errstate(invalid='raise'):
        try:
            B = np.true_divide(B, absB)
        except FloatingPointError as e:
            errargs = np.argwhere(absB == 0)
            # Report error where there is zero divides for a non-zero freq
            if not np.all(errargs[:, 0] == 0):
                logging.debug(f'{e} {errargs}')
    # Set zero frequency component to zero
    B[:, 0] = 0.j

    return B



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Rounds to next 0.5

Created on Sat May  2 13:48:00 2020

Author:
    Peter Makus (peter.makus@student.uib.no)

Last updated:
"""


def roundhalf(number: float) -> float:
    """
    Rounds to next half of integer

    Parameters
    ----------
    number : float/int
        number to be rounded.

    Returns
    -------
    float
        Closest half of integer.

    """
    return round(number*2) / 2



'''
Plotting specral time series.

:copyright:
   The SeisMIC development team (makus@gfz-potsdam.de).
:license:
   EUROPEAN UNION PUBLIC LICENCE Version 1.2
   (https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12)
:author:
   Peter Makus (makus@gfz-potsdam.de)

Created: Wednesday, 21st June 2023 04:54:20 pm
Last Modified: Monday, 21st August 2023 12:22:56 pm
'''

import os
from datetime import datetime
from typing import Tuple
import locale

import numpy as np
from matplotlib import pyplot as plt
from matplotlib import dates as mdates
from matplotlib import colors
from obspy import UTCDateTime

from seismic.plot.plot_utils import set_mpl_params

# Batlow colour scale
batlow = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), 'batlow.txt')


def plot_spct_series(
    S: np.ndarray, f: np.ndarray, t: np.ndarray, norm: str = None,
    norm_method: str = None, flim: Tuple[int, int] = None,
    tlim: Tuple[datetime, datetime] = None, cmap: str = 'batlow', vmin=None,
        vmax=None, log_scale: bool = False) -> plt.Axes:
    """
    Plots a spectral series.

    :param S: A spectral series with dim=2.
    :type S: np.ndarray
    :param f: Vector containing frequencies (Hz)
    :type f: np.ndarray
    :param t: Vector containing times as UTCDateTimes
    :type t: np.ndarray
    :param norm: Normalise the spectrum either on the time axis with
        norm='t' or on the frequency axis with norm='f', defaults to None.
    :type norm: str, optional
    :param norm_method: Normation method to use.
        Either 'linalg' (i.e., length of vector),
        'mean', or 'median'.
    :param title: Plot title, defaults to None
    :type title: str, optional
    :param outfile: location to save the plot to, defaults to None
    :type outfile: str, optional
    :param fmt: Format to save figure, defaults to 'pdf'
    :type fmt: str, optional
    :param flim: Limit Frequency axis and Normalisation to the values
        in the given window
    :type flim: Tuple[int, int]
    :param tlim: Limit time axis to the values in the given window
    :type tlim: Tuple[datetime, datetime]
    """
    # Mask nans
    S = np.ma.masked_invalid(S)

    # Show dates in English format
    locale.setlocale(locale.LC_ALL, "en_GB.utf8")
    # Create UTC time series

    set_mpl_params()

    if log_scale:
        plt.yscale('log')

    if flim is not None:
        plt.ylim(flim)
        ii = np.argmin(abs(f-flim[0]))
        jj = np.argmin(abs(f-flim[1])) + 1
        f = f[ii:jj]
        S = S[ii:jj, :]
    else:
        plt.ylim(.1, f.max())

    t = np.array([UTCDateTime(tt).datetime for tt in t])

    if tlim is not None:
        plt.xlim(tlim)
        ii = np.argmin(abs(t-tlim[0]))
        jj = np.argmin(abs(t-tlim[1]))
        t = t[ii:jj]
        S = S[:, ii:jj]

    # Normalise
    if not norm:
        pass
    elif norm == 'f':
        if norm_method == 'linalg':
            S = np.divide(S, np.linalg.norm(S, axis=1)[:, np.newaxis])
        elif norm_method == 'mean':
            S = np.divide(S, np.mean(S, axis=1)[:, np.newaxis])
        elif norm_method == 'median':
            S = np.divide(S, np.median(S, axis=1)[:, np.newaxis])
        else:
            raise ValueError('Normalisation method %s unkown.' % norm_method)
    elif norm == 't':
        if norm_method == 'linalg':
            S = np.divide(S, np.linalg.norm(S, axis=0))
        elif norm_method == 'mean':
            S = np.divide(S, np.mean(S, axis=0))
        elif norm_method == 'median':
            S = np.divide(S, np.median(S, axis=0))
        else:
            raise ValueError('Normalisation method %s unkown.' % norm_method)
    else:
        raise ValueError('Normalisation %s unkown.' % norm)
    if cmap == 'batlow':
        cmap = colors.LinearSegmentedColormap.from_list(
            'batlow', np.loadtxt(batlow))
        cmap.set_bad('k')
    S /= S.max()
    pcm = plt.pcolormesh(
        t, f, S, shading='gouraud', cmap=cmap,
        norm=colors.LogNorm(), vmin=vmin, vmax=vmax)
    plt.colorbar(
        pcm, label='Energy (normalised)', orientation='horizontal', shrink=.6,
        pad=.25)
    plt.ylabel(r'$f$ [Hz]')
    ax = plt.gca()
    ax.xaxis.set_major_locator(mdates.AutoDateLocator())
    ax.xaxis.set_major_formatter(
        mdates.AutoDateFormatter(ax.get_axes_locator()))
    plt.xticks(rotation=35)
    return ax
